from django.conf import settings
from django.utils.module_loading import import_string


def link_components(request, user=None):
    protocol = 'https' if request.is_secure() else 'http'
    hostname = request.get_host()

    return {"protocol": protocol, "hostname": hostname}


def build_email_context(request=None, user=None):
    context = {'user': user}

    processors = []
    collect = []
    collect.extend(getattr(settings, "EMAIL_CONTEXT_PROCESSORS",
                           ('smartmin.email.link_components',)))
    for path in collect:
        func = import_string(path)
        processors.append(func)

    for processor in processors:
        context.update(processor(request, user))

    return context



from django.test import TestCase

from .models import ImportTask, generate_file_path


class ImportTest(TestCase):
    def test_csv_import(self):
        pass

    def test_generate_file_path(self):
        self.assertEquals(generate_file_path(ImportTask(), 'allo.csv'), 'csv_imports/allo.csv')
        self.assertEquals(generate_file_path(ImportTask(), 'allo.xlsx'), 'csv_imports/allo.xlsx')
        self.assertEquals(generate_file_path(ImportTask(), 'allo.foo.bar'), 'csv_imports/allo.foo.bar')

        long_name = 'foo' * 100

        test_file_name = '%s.xls.csv' % long_name
        self.assertEquals(len(generate_file_path(ImportTask(), test_file_name)), 100)
        self.assertEquals(generate_file_path(ImportTask(), test_file_name), 'csv_imports/%s.csv' % long_name[:84])

        test_file_name = '%s.abc.xlsx' % long_name
        self.assertEquals(len(generate_file_path(ImportTask(), test_file_name)), 100)
        self.assertEquals(generate_file_path(ImportTask(), test_file_name), 'csv_imports/%s.xlsx' % long_name[:83])



import os

from django.db import models
from django.utils import timezone
from django.utils.module_loading import import_string

from smartmin.models import SmartModel


def generate_file_path(instance, filename):

    file_path_prefix = 'csv_imports/'

    name, extension = os.path.splitext(filename)

    if len(name) + len(extension) >= 100:
        name = name[:100-len(extension)-len(file_path_prefix)]

    return "%s%s%s" % (file_path_prefix, name, extension)


class ImportTask(SmartModel):
    PENDING = 'PENDING'
    STARTED = 'STARTED'
    RUNNING = 'RUNNING'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'

    READY_STATES = [SUCCESS, FAILURE]

    csv_file = models.FileField(upload_to=generate_file_path, verbose_name="Import file",
                                help_text="A comma delimited file of records to import")

    model_class = models.CharField(max_length=255, help_text="The model we are importing for")

    import_params = models.TextField(blank=True, null=True, help_text="JSON blob of form parameters on task creation")

    import_log = models.TextField()

    import_results = models.TextField(blank=True, null=True, help_text="JSON blob of result values on task completion")

    task_id = models.CharField(null=True, max_length=64)

    task_status = models.CharField(max_length=32, default=PENDING)

    def start(self):
        from .tasks import csv_import
        self.log("Queued import at %s" % timezone.now())
        self.task_status = self.STARTED
        self.save(update_fields=['import_log', 'task_status'])
        result = csv_import.delay(self.pk)
        self.task_id = result.task_id
        self.save(update_fields=['task_id'])

    def done(self):
        if self.task_id:
            return self.task_status in self.READY_STATES

    def status(self):
        return self.task_status

    def log(self, message):
        self.import_log += "%s\n" % message
        self.modified_on = timezone.now()
        self.save(update_fields=['import_log', 'modified_on'])

    def __unicode__(self):
        return "%s Import" % import_string(self.model_class)._meta.verbose_name.title()



# Generated by Django 2.2.17 on 2021-02-19 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0002_remove_failed_logins'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='failedlogin',
            name='user',
        ),
        migrations.AddField(
            model_name='failedlogin',
            name='username',
            field=models.CharField(default=None, max_length=256),
            preserve_default=False,
        ),
    ]



import random
import string
from datetime import timedelta

from django import forms
from django.conf import settings
from django.contrib import messages, auth
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.auth.views import LoginView
from django.core.mail import send_mail
from django.urls import reverse
from django.http import HttpResponseRedirect
from django.template import loader
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.views.generic import TemplateView

from smartmin.email import build_email_context
from smartmin.views import SmartCRUDL, SmartView, SmartFormView, SmartListView, SmartCreateView, SmartUpdateView
from .models import RecoveryToken, PasswordHistory, FailedLogin, is_password_complex


class UserForm(forms.ModelForm):
    new_password = forms.CharField(label=_("New Password"), widget=forms.PasswordInput, strip=False)
    groups = forms.ModelMultipleChoiceField(widget=forms.CheckboxSelectMultiple,
                                            queryset=Group.objects.all(), required=False)

    def clean_new_password(self):
        password = self.cleaned_data['new_password']

        # if they specified a new password
        if password and not is_password_complex(password):
            raise forms.ValidationError(_("Passwords must have at least 8 characters, including one uppercase, "
                                          "one lowercase and one number"))

        return password

    def save(self, commit=True):
        """
        Overloaded so we can save any new password that is included.
        """
        is_new_user = self.instance.pk is None

        user = super(UserForm, self).save(commit)

        # new users should be made active by default
        if is_new_user:
            user.is_active = True

        # if we had a new password set, use it
        new_pass = self.cleaned_data['new_password']
        if new_pass:
            user.set_password(new_pass)
            if commit:
                user.save()

        return user

    class Meta:
        model = get_user_model()
        fields = ('username', 'new_password', 'first_name', 'last_name', 'email', 'groups', 'is_active')


class UserUpdateForm(UserForm):
    new_password = forms.CharField(label=_("New Password"), widget=forms.PasswordInput, required=False, strip=False)

    def clean_new_password(self):
        password = self.cleaned_data['new_password']

        if password and not is_password_complex(password):
            raise forms.ValidationError(_("Passwords must have at least 8 characters, including one uppercase, "
                                          "one lowercase and one number"))

        if password and PasswordHistory.is_password_repeat(self.instance, password):
            raise forms.ValidationError(_("You have used this password before in the past year, "
                                          "please use a new password."))

        return password


class UserProfileForm(UserForm):
    old_password = forms.CharField(label=_("Password"), widget=forms.PasswordInput, required=False, strip=False)
    new_password = forms.CharField(label=_("New Password"), widget=forms.PasswordInput, required=False, strip=False)
    confirm_new_password = forms.CharField(
        label=_("Confirm Password"), widget=forms.PasswordInput, required=False, strip=False
    )

    def clean_old_password(self):
        user = self.instance

        if(not user.check_password(self.cleaned_data['old_password'])):
            raise forms.ValidationError(_("Please enter your password to save changes."))

        return self.cleaned_data['old_password']

    def clean_confirm_new_password(self):
        if 'new_password' not in self.cleaned_data:
            return None

        if not self.cleaned_data['confirm_new_password'] and self.cleaned_data['new_password']:
            raise forms.ValidationError(_("Confirm the new password by filling the this field"))

        if self.cleaned_data['new_password'] != self.cleaned_data['confirm_new_password']:
            raise forms.ValidationError(_("New password doesn't match with its confirmation"))

        password = self.cleaned_data['new_password']
        if password and not is_password_complex(password):
            raise forms.ValidationError(_("Passwords must have at least 8 characters, including one uppercase, "
                                          "one lowercase and one number"))

        if password and PasswordHistory.is_password_repeat(self.instance, password):
            raise forms.ValidationError(_("You have used this password before in the past year, "
                                          "please use a new password."))

        return self.cleaned_data['new_password']


class UserForgetForm(forms.Form):
    email = forms.EmailField(label=_("Your Email"),)

    def clean_email(self):
        email = self.cleaned_data['email'].strip()

        allow_email_recovery = getattr(settings, 'USER_ALLOW_EMAIL_RECOVERY', True)
        if not allow_email_recovery:
            raise forms.ValidationError(_("E-mail recovery is not supported, "
                                          "please contact the website administrator to reset your password manually."))

        return email


class SetPasswordForm(UserForm):
    old_password = forms.CharField(label=_("Current Password"), widget=forms.PasswordInput, required=True, strip=False,
                                   help_text=_("Your current password"))
    new_password = forms.CharField(label=_("New Password"), widget=forms.PasswordInput, required=True,
                                   help_text=_("Your new password."), strip=False)
    confirm_new_password = forms.CharField(label=_("Confirm new Password"), widget=forms.PasswordInput, required=True,
                                           help_text=_("Confirm your new password."), strip=False)

    def clean_old_password(self):
        user = self.instance
        if not user.check_password(self.cleaned_data['old_password']):
            raise forms.ValidationError(_("Please enter your password to save changes"))

        return self.cleaned_data['old_password']

    def clean_confirm_new_password(self):
        if 'new_password' not in self.cleaned_data:
            return None

        if not self.cleaned_data['confirm_new_password'] and self.cleaned_data['new_password']:
            raise forms.ValidationError(_("Confirm your new password by entering it here"))

        if self.cleaned_data['new_password'] != self.cleaned_data['confirm_new_password']:
            raise forms.ValidationError(_("Mismatch between your new password and confirmation, try again"))

        password = self.cleaned_data['new_password']
        if password and not is_password_complex(password):
            raise forms.ValidationError(_("Passwords must have at least 8 characters, including one uppercase, "
                                          "one lowercase and one number"))

        if password and PasswordHistory.is_password_repeat(self.instance, password):
            raise forms.ValidationError(_("You have used this password before in the past year, "
                                          "please use a new password."))

        return self.cleaned_data['new_password']


class UserCRUDL(SmartCRUDL):
    model = get_user_model()
    permissions = True
    actions = ('create', 'list', 'update', 'profile', 'forget', 'recover', 'expired', 'failed', 'newpassword', 'mimic')

    class List(SmartListView):
        search_fields = ('username__icontains', 'first_name__icontains', 'last_name__icontains')
        fields = ('username', 'name', 'group', 'last_login')
        link_fields = ('username', 'name')
        default_order = 'username'
        add_button = True
        template_name = "smartmin/users/user_list.html"

        def get_context_data(self, **kwargs):
            context = super(UserCRUDL.List, self).get_context_data(**kwargs)
            context['groups'] = Group.objects.all()
            group_id = self.request.POST.get('group_id', self.request.GET.get('group_id', 0))
            context['group_id'] = int(group_id)
            return context

        def get_group(self, obj):
            return ", ".join([group.name for group in obj.groups.all()])

        def get_queryset(self, **kwargs):
            queryset = super(UserCRUDL.List, self).get_queryset(**kwargs)
            group_id = self.request.POST.get('group_id', self.request.GET.get('group_id', 0))
            group_id = int(group_id)

            # filter by the group
            if group_id:
                queryset = queryset.filter(groups=group_id)

            # ignore superusers and staff users
            return queryset.exclude(is_staff=True).exclude(is_superuser=True).exclude(password=None)

        def get_name(self, obj):
            return obj.get_full_name()

    class Create(SmartCreateView):
        form_class = UserForm
        fields = ('username', 'new_password', 'first_name', 'last_name', 'email', 'groups')
        success_message = _("New user created successfully.")

        field_config = {
            'groups': dict(label=_("Groups"),
                           help=_("Users will only get those permissions that are allowed for their group.")),
            'new_password': dict(label=_("Password"), help=_("Set the user's initial password here.")),
        }

        def post_save(self, obj):
            """
            Make sure our groups are up to date
            """
            if 'groups' in self.form.cleaned_data:
                for group in self.form.cleaned_data['groups']:
                    obj.groups.add(group)

            return obj

    class Update(SmartUpdateView):
        form_class = UserUpdateForm
        template_name = "smartmin/users/user_update.html"
        success_message = "User saved successfully."
        fields = ('username', 'new_password', 'first_name', 'last_name', 'email', 'groups', 'is_active', 'last_login')
        field_config = {
            'last_login': dict(readonly=True, label=_("Last Login")),
            'is_active': dict(label=_("Is Active"), help=_("Whether this user is allowed to log into the site")),
            'groups': dict(label=_("Groups"),
                           help=_("Users will only get those permissions that are allowed for their group")),
            'new_password': dict(label=_("New Password"),
                                 help=_("You can reset the user's password by entering a new password here")),
        }

        def post_save(self, obj):
            """
            Make sure our groups are up to date
            """
            if 'groups' in self.form.cleaned_data:
                obj.groups.clear()
                for group in self.form.cleaned_data['groups']:
                    obj.groups.add(group)

            # if a new password was set, reset our failed logins
            if 'new_password' in self.form.cleaned_data and self.form.cleaned_data['new_password']:
                FailedLogin.objects.filter(username__iexact=self.object.username).delete()
                PasswordHistory.objects.create(user=obj, password=obj.password)

            return obj

    class Profile(SmartUpdateView):
        form_class = UserProfileForm
        success_message = "User profile saved successfully."
        fields = ('username', 'old_password', 'new_password', 'confirm_new_password',
                  'first_name', 'last_name', 'email')
        field_config = {
            'username': dict(readonly=True, label=_("Username")),
            'old_password': dict(label=_("Password"), help=_("Your password")),
            'new_password': dict(label=_("New Password"), help=_("If you want to set a new password, enter it here")),
            'confirm_new_password': dict(label=_("Confirm New Password"), help=_("Confirm your new password")),
        }

        def post_save(self, obj):
            obj = super(UserCRUDL.Profile, self).post_save(obj)
            if 'new_password' in self.form.cleaned_data and self.form.cleaned_data['new_password']:
                FailedLogin.objects.filter(username__iexact=self.object.username).delete()
                PasswordHistory.objects.create(user=obj, password=obj.password)

            return obj

        def get_object(self, queryset=None):
            return self.request.user

        def derive_title(self):
            return _("Edit your profile")

    class Forget(SmartFormView):
        title = _("Password Recovery")
        template_name = 'smartmin/users/user_forget.html'
        form_class = UserForgetForm
        permission = None
        success_message = _("An Email has been sent to your account with further instructions.")
        success_url = "@users.user_login"
        fields = ('email', )

        def form_valid(self, form):
            email = form.cleaned_data['email']
            hostname = getattr(settings, 'HOSTNAME', self.request.get_host())

            col_index = hostname.find(':')
            domain = hostname[:col_index] if col_index > 0 else hostname

            from_email = getattr(settings, 'DEFAULT_FROM_EMAIL', 'website@%s' % domain)
            user_email_template = getattr(settings, "USER_FORGET_EMAIL_TEMPLATE", "smartmin/users/user_email.txt")

            user = get_user_model().objects.filter(email__iexact=email).first()

            context = build_email_context(self.request, user)

            if user:
                token = ''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(32))
                RecoveryToken.objects.create(token=token, user=user)
                email_template = loader.get_template(user_email_template)
                FailedLogin.objects.filter(username__iexact=user.username).delete()
                context['user'] = user
                context['path'] = "%s" % reverse('users.user_recover', args=[token])

                send_mail(_('Password Recovery Request'), email_template.render(context), from_email,
                            [email], fail_silently=False)

            response = super(UserCRUDL.Forget, self).form_valid(form)
            return response

    class Newpassword(SmartUpdateView):
        form_class = SetPasswordForm
        fields = ('old_password', 'new_password', 'confirm_new_password')
        title = _("Pick a new password")
        template_name = 'smartmin/users/user_newpassword.html'
        success_message = _("Your password has successfully been updated, thank you.")

        def get_context_data(self, *args, **kwargs):
            context_data = super(UserCRUDL.Newpassword, self).get_context_data(*args, **kwargs)
            context_data['expire_days'] = getattr(settings, 'USER_PASSWORD_EXPIRATION', -1)
            context_data['window_days'] = getattr(settings, 'USER_PASSWORD_REPEAT_WINDOW', -1)
            return context_data

        def has_permission(self, request, *args, **kwargs):
            return request.user.is_authenticated

        def get_object(self, queryset=None):
            return self.request.user

        def post_save(self, obj):
            obj = super(UserCRUDL.Newpassword, self).post_save(obj)
            PasswordHistory.objects.create(user=obj, password=obj.password)
            return obj

        def get_success_url(self):
            return settings.LOGIN_REDIRECT_URL

    class Mimic(SmartUpdateView):
        fields = ('id',)

        def derive_success_message(self):
            return _("You are now logged in as %s") % self.object.username

        def pre_process(self, request, *args, **kwargs):
            user = self.get_object()

            Login.as_view()(request)

            # After logging in it is important to change the user stored in the session
            # otherwise the user will remain the same
            request.session[auth.SESSION_KEY] = user.id
            request.session[auth.HASH_SESSION_KEY] = user.get_session_auth_hash()

            return HttpResponseRedirect(settings.LOGIN_REDIRECT_URL)

    class Recover(SmartUpdateView):
        form_class = SetPasswordForm
        permission = None
        success_message = _("Password Updated Successfully. Now you can log in using your new password.")
        success_url = '@users.user_login'
        fields = ('new_password', 'confirm_new_password')
        title = _("Reset your Password")
        template_name = 'smartmin/users/user_recover.html'

        @classmethod
        def derive_url_pattern(cls, path, action):
            return r'^%s/%s/(?P<token>\w+)/$' % (path, action)

        def pre_process(self, request, *args, **kwargs):
            token = self.kwargs.get('token')
            validity_time = timezone.now() - timedelta(hours=48)
            recovery_token = RecoveryToken.objects.filter(created_on__gt=validity_time, token=token)
            if not recovery_token:
                messages.info(request, _("Your link has expired for security reasons. "
                                         "Please reinitiate the process by entering your email here."))
                return HttpResponseRedirect(reverse("users.user_forget"))
            return super(UserCRUDL.Recover, self).pre_process(request, args, kwargs)

        def get_object(self, queryset=None):
            token = self.kwargs.get('token')
            recovery_token = RecoveryToken.objects.get(token=token)
            return recovery_token.user

        def post_save(self, obj):
            obj = super(UserCRUDL.Recover, self).post_save(obj)
            validity_time = timezone.now() - timedelta(hours=48)
            RecoveryToken.objects.filter(user=obj).delete()
            RecoveryToken.objects.filter(created_on__lt=validity_time).delete()
            PasswordHistory.objects.create(user=obj, password=obj.password)
            return obj

    class Expired(SmartView, TemplateView):
        permission = None
        template_name = 'smartmin/users/user_expired.html'

    class Failed(SmartView, TemplateView):
        permission = None
        template_name = 'smartmin/users/user_failed.html'

        def get_context_data(self, *args, **kwargs):
            context = super(UserCRUDL.Failed, self).get_context_data(*args, **kwargs)

            lockout_timeout = getattr(settings, 'USER_LOCKOUT_TIMEOUT', 10)
            failed_login_limit = getattr(settings, 'USER_FAILED_LOGIN_LIMIT', 5)
            allow_email_recovery = getattr(settings, 'USER_ALLOW_EMAIL_RECOVERY', True)

            context['lockout_timeout'] = lockout_timeout
            context['failed_login_limit'] = failed_login_limit
            context['allow_email_recovery'] = allow_email_recovery

            return context


class Login(LoginView):
    template_name = 'smartmin/users/login.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['allow_email_recovery'] = getattr(settings, 'USER_ALLOW_EMAIL_RECOVERY', True)

        return context

    def post(self, request, *args, **kwargs):
        form = self.get_form()

        # clean form data
        form_is_valid = form.is_valid()

        lockout_timeout = getattr(settings, 'USER_LOCKOUT_TIMEOUT', 10)
        failed_login_limit = getattr(settings, 'USER_FAILED_LOGIN_LIMIT', 5)

        username = self.get_username(form)
        if not username:
            return self.form_invalid(form)

        user = get_user_model().objects.filter(username__iexact=username).first()
        valid_password = False
        
        # this could be a valid login by a user
        if user:
            # incorrect password?  create a failed login token
            valid_password = user.check_password(form.cleaned_data.get('password'))
            
        if not user or not valid_password:
            FailedLogin.objects.create(username=username)

        bad_interval = timezone.now() - timedelta(minutes=lockout_timeout)
        failures = FailedLogin.objects.filter(username__iexact=username)

        # if the failures reset after a period of time, then limit our query to that interval
        if lockout_timeout > 0:
            failures = failures.filter(failed_on__gt=bad_interval)

        # if there are too many failed logins, take them to the failed page
        if len(failures) >= failed_login_limit:
            return HttpResponseRedirect(reverse('users.user_failed'))

        # pass through the normal login process
        if form_is_valid:
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        # clean up any failed logins for this user
        FailedLogin.objects.filter(username__iexact=self.get_username(form)).delete()

        return super().form_valid(form)

    def get_username(self, form):
        return form.cleaned_data.get('username')



import time
from importlib import import_module

from django.core.management.base import BaseCommand, CommandError
from django.core.management.sql import emit_pre_migrate_signal, emit_post_migrate_signal
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.loader import AmbiguityError

APPLY_FUNCTION = 'apply_manual'


class Command(BaseCommand):  # pragma: no cover
    help = "Applies a migration manually which may have been previously applied or faked"

    def add_arguments(self, parser):
        parser.add_argument('app_label',
                            help='App label of an application to synchronize the state.')
        parser.add_argument('migration_name',
                            help='Database state will be brought to the state after that migration.')
        parser.add_argument('--record', action='store_true', dest='record', default=False,
                            help='Record migration as applied.')

    def handle(self, app_label, migration_name, *args, **options):
        self.verbosity = options.get('verbosity')
        interactive = options.get('interactive')
        record = options.get('record')

        connection = connections[DEFAULT_DB_ALIAS]
        connection.prepare_database()
        executor = MigrationExecutor(connection, None)

        # before anything else, see if there's conflicting apps and drop out hard if there are any
        conflicts = executor.loader.detect_conflicts()
        if conflicts:
            name_str = "; ".join("%s in %s" % (", ".join(names), app) for app, names in conflicts.items())
            raise CommandError(
                "Conflicting migrations detected (%s).\nTo fix them run "
                "'python manage.py makemigrations --merge'" % name_str
            )

        if app_label not in executor.loader.migrated_apps:
            raise CommandError(
                "App '%s' does not have migrations (you cannot selectively sync unmigrated apps)" % app_label
            )
        try:
            migration = executor.loader.get_migration_by_prefix(app_label, migration_name)
        except AmbiguityError:
            raise CommandError(
                "More than one migration matches '%s' in app '%s'. Please be more specific." %
                (migration_name, app_label)
            )
        except KeyError:
            raise CommandError("Cannot find a migration matching '%s' from app '%s'." % (migration_name, app_label))

        migration_module = import_module(migration.__module__)

        # check migration can be run offline
        apply_function = getattr(migration_module, APPLY_FUNCTION, None)
        if not apply_function or not callable(apply_function):
            raise CommandError("Migration %s does not contain function named '%s'" % (migration, APPLY_FUNCTION))

        plan = executor.migration_plan([(app_label, migration.name)])
        if record and not plan:
            raise CommandError("Migration %s has already been applied" % migration)

        emit_pre_migrate_signal(self.verbosity, interactive, connection.alias)

        self.stdout.write(self.style.MIGRATE_HEADING("Operations to perform:"))
        self.stdout.write("  Manually apply migration %s" % migration)
        if record:
            self.stdout.write("  Record migration %s as applied" % migration)
        self.stdout.write(self.style.MIGRATE_HEADING("Manual migration:"))

        self.apply_migration(migration, apply_function)

        if record:
            self.record_migration(migration, executor)

        # send the post_migrate signal, so individual apps can do whatever they need to do at this point.
        emit_post_migrate_signal(self.verbosity, interactive, connection.alias)

    def apply_migration(self, migration, apply_function):
        compute_time = self.verbosity > 1

        self.stdout.write("  Applying %s... " % migration, ending="")
        self.stdout.flush()

        start = time.time()

        apply_function()

        elapsed = " (%.3fs)" % (time.time() - start) if compute_time else ""

        self.stdout.write(self.style.SUCCESS("OK" + elapsed))

    def record_migration(self, migration, executor):
        self.stdout.write("  Recording %s... " % migration, ending="")
        self.stdout.flush()

        executor.recorder.record_applied(migration.app_label, migration.name)

        self.stdout.write(self.style.SUCCESS("DONE"))




from urllib.parse import urlparse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.urls import reverse
from django.test.testcases import TestCase
from django.utils.encoding import force_str


class SmartminTestMixin(object):

    def fetch_protected(self, url, user, post_data=None, failOnFormValidation=True):
        """
        Fetches the given url. Fails if it can be fetched without first logging in as given user
        """
        # make sure we are logged out before testing permissions
        self.client.logout()

        # can't load if we aren't logged in
        response = self.client.get(url)
        self.assertRedirect(response, reverse("users.user_login"),
                            msg="'%s' loaded without being logged in first" % url)
        self.login(user)

        # but now we can!
        if not post_data:
            response = self.client.get(url)
            self.assertEquals(200, response.status_code)
        else:
            response = self.client.post(url, data=post_data)
            self.assertNotRedirect(response, reverse("users.user_login"), msg="Unexpected redirect to login")

            if failOnFormValidation:
                self.assertNoFormErrors(response, post_data)
                self.assertEquals(302, response.status_code)

        return response

    def assertLoginRedirect(self, response, msg=None):
        self.assertRedirect(response, settings.LOGIN_URL, msg=msg)

    def assertRedirect(self, response, url, status_code=302, msg=None):
        self.assertEqual(response.status_code, status_code, msg=msg)
        segments = urlparse(response.get('Location', None))
        self.assertEqual(segments.path, url, msg=msg)

    def assertNotRedirect(self, response, url, msg=None):
        if response.status_code == 302:
            segments = urlparse(response.get('Location', None))
            self.assertNotEqual(segments.path, url, msg=msg)

    def create_user(self, username, group_names=()):
        # Create a user to run our CRUDL tests
        user = get_user_model().objects.create_user(username, "%s@nyaruka.com" % username)
        user.set_password(username)
        user.save()
        for group in group_names:
            user.groups.add(Group.objects.get(name=group))
        return user

    def login(self, user):
        self.assertTrue(self.client.login(username=user.username, password=user.username),
                        "Couldn't login as %(user)s / %(user)s" % dict(user=user.username))

    def assertNoFormErrors(self, response, post_data=None):
        if response.status_code == 200 and 'form' in response.context:
            form = response.context['form']

            if not form.is_valid():
                errors = []
                for k, v in form.errors.items():
                    errors.append("%s=%s" % (k, force_str(v)))
                self.fail("Create failed with form errors: %s, Posted: %s" % (",".join(errors), post_data))

    def create_anonymous_user(self):
        User = get_user_model()
        _anon_exists = User.objects.filter(username=settings.ANONYMOUS_USER_NAME).exists()
        if not _anon_exists:
            user = User(username=settings.ANONYMOUS_USER_NAME)
            user.set_unusable_password()
            user.save()


class SmartminTest(SmartminTestMixin, TestCase):
    pass


class _CRUDLTest(SmartminTest):
    """
    Base class for standard CRUDL test cases
    """
    crudl = None
    user = None
    object = None

    def setUp(self):
        self.crudl = None
        self.user = None
        super(_CRUDLTest, self).setUp()

    def run(self, result=None):
        # only actually run sub classes of this
        if self.__class__ != _CRUDLTest:
            super(_CRUDLTest, self).run(result)

    def getCRUDL(self):
        if self.crudl:
            return self.crudl()
        raise Exception("Must define self.crudl")

    def getUser(self):
        if self.user:
            return self.user
        raise Exception("Must define self.user")

    def getCreatePostData(self):
        raise Exception("Missing method: %s.getCreatePostData()" % self.__class__.__name__)

    def getUpdatePostData(self):
        raise Exception("Missing method: %s.getUpdatePostData()" % self.__class__.__name__)

    def getManager(self):
        return self.getCRUDL().model.objects

    def getTestObject(self):
        if self.object:
            return self.object

        if self.getCRUDL().permissions:
            self.login(self.getUser())

        # create our object
        create_page = reverse(self.getCRUDL().url_name_for_action('create'))
        post_data = self.getCreatePostData()
        self.client.post(create_page, data=post_data)

        # find our created object
        self.object = self.getManager().get(**post_data)
        return self.object

    def testCreate(self):
        if 'create' not in self.getCRUDL().actions:
            return
        self._do_test_view('create', post_data=self.getCreatePostData())

    def testRead(self):
        if 'read' not in self.getCRUDL().actions:
            return
        self._do_test_view('read', self.getTestObject())

    def testUpdate(self):
        if 'update' not in self.getCRUDL().actions:
            return
        self._do_test_view('update', self.getTestObject(), post_data=self.getUpdatePostData())

    def testDelete(self):
        if 'delete' not in self.getCRUDL().actions:
            return
        object = self.getTestObject()
        self._do_test_view('delete', object, post_data=dict())
        self.assertEquals(0, len(self.getManager().filter(pk=object.pk)))

    def testList(self):
        if 'list' not in self.getCRUDL().actions:
            return
        # have at least one object
        self.getTestObject()
        self._do_test_view('list')

    def testCsv(self):
        if 'csv' not in self.getCRUDL().actions:
            return
        # have at least one object
        self.getTestObject()
        self._do_test_view('csv')

    def _do_test_view(self, action=None, object=None, post_data=None, query_string=None):
        url_name = self.getCRUDL().url_name_for_action(action)
        if object:
            url = reverse(url_name, args=[object.pk])
        else:
            url = reverse(url_name)

        # append our query string if we have one
        if query_string:
            url = "%s?%s" % (url, query_string)

        # make sure we are logged out before testing permissions
        self.client.logout()

        response = self.client.get(url)

        view = self.getCRUDL().view_for_action(action)
        if self.getCRUDL().permissions and view.permission is not None:
            self.assertRedirect(response, reverse("users.user_login"),
                                msg="Page for '%s' loaded without being logged in first" % action)
            self.login(self.getUser())
            response = self.client.get(url)

        fn = "assert%sGet" % action.capitalize()
        self.assertPageGet(action, response)
        if hasattr(self, fn):
            getattr(self, fn)(response)

        if post_data is not None:
            response = self.client.post(url, data=post_data)

            self.assertPagePost(action, response)
            fn = "assert%sPost" % action.capitalize()
            if hasattr(self, fn):
                getattr(self, fn)(response)

        return response

    def assertPageGet(self, action, response):
        if response.status_code == 302:
            self.fail("'%s' resulted in an unexpected redirect to: %s" % (action, response.get('Location')))
        self.assertEquals(200, response.status_code, )

    def assertPagePost(self, action, response):
        self.assertNoFormErrors(response)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend


class CaseInsensitiveBackend(ModelBackend):
    """
    Authenticates against settings.AUTH_USER_MODEL.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        User = get_user_model()
        try:
            user = User.objects.get(username__iexact=username)
            if user.check_password(password):
                return user
            else:
                return None
        except User.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a non-existing user (#20760).
            User().set_password(password)





import json
import operator
from functools import reduce

import django.forms.models as model_forms
from django import forms
from django.conf import settings
from django.conf.urls import url
from django.contrib import messages
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.db import IntegrityError
from django.db.models import Q
from django.http import HttpResponseRedirect, HttpResponse, JsonResponse
from django.utils.encoding import force_text
from django.utils.http import urlquote
from django.utils.translation import ugettext_lazy as _
from django.views.generic.edit import ModelFormMixin, UpdateView, CreateView, ProcessFormView, FormView
from django.views.generic.base import TemplateView
from django.views.generic import DetailView, ListView

from smartmin.csv_imports.models import ImportTask
from smartmin.mixins import NonAtomicMixin

from . import widgets


def smart_url(url, obj=None):
    """
    URLs that start with @ are reversed, using the passed in arguments.

    Otherwise a straight % substitution is applied.
    """
    if url.find("@") >= 0:
        (args, value) = url.split('@')

        if args:
            val = getattr(obj, args, None)
            return reverse(value, args=[val])
        else:
            return reverse(value)
    else:
        if obj is None:
            return url
        else:
            return url % obj.id


class SmartView:
    fields = None
    exclude = None
    field_config = {}
    title = None
    refresh = 0
    template_name = None
    pjax = None

    # set by our CRUDL
    url_name = None

    # if we are part of a CRUDL, we keep a reference to it here, set by CRUDL
    crudl = None

    def __init__(self, *args):
        """
        There are a few variables we want to mantain in the instance, not the
        class.
        """
        self.extra_context = {}
        super(SmartView, self).__init__()

    def derive_title(self):
        """
        Returns the title used on this page.
        """
        return self.title

    @classmethod
    def derive_url_pattern(cls, path, action):
        """
        Returns the URL pattern for this view.
        """
        return r'^%s/%s/$' % (path, action)

    def has_permission(self, request, *args, **kwargs):
        """
        Figures out if the current user has permissions for this view.
        """
        self.kwargs = kwargs
        self.args = args
        self.request = request

        if not getattr(self, 'permission', None):
            return True
        else:
            return request.user.has_perm(self.permission)

    def dispatch(self, request, *args, **kwargs):
        """
        Overloaded to check permissions if appropriate
        """
        def wrapper(request, *args, **kwargs):
            if not self.has_permission(request, *args, **kwargs):
                path = urlquote(request.get_full_path())
                login_url = kwargs.pop('login_url', settings.LOGIN_URL)
                redirect_field_name = kwargs.pop('redirect_field_name', REDIRECT_FIELD_NAME)
                return HttpResponseRedirect("%s?%s=%s" % (login_url, redirect_field_name, path))
            else:
                response = self.pre_process(request, *args, **kwargs)
                if not response:
                    return super(SmartView, self).dispatch(request, *args, **kwargs)
                else:
                    return response

        return wrapper(request, *args, **kwargs)

    def pre_process(self, request, *args, **kwargs):
        """
        Gives the view an opportunity to intercept this response and return a different
        response instead.  This can be used to check some precondition for example and to
        redirect the user somewhere else if they are not met.

        Views which wish to use this should return a Response object.
        """
        return None

    def lookup_obj_attribute(self, obj, field):
        """
        Looks for a field's value from the passed in obj.  Note that this will strip
        leading attributes to deal with subelements if possible
        """
        curr_field = field.encode('ascii', 'ignore').decode("utf-8")
        rest = None

        if field.find('.') >= 0:
            curr_field = field.split('.')[0]
            rest = '.'.join(field.split('.')[1:])

        # next up is the object itself
        obj_field = getattr(obj, curr_field, None)

        # if it is callable, do so
        if obj_field and getattr(obj_field, '__call__', None):
            obj_field = obj_field()

        if obj_field and rest:
            return self.lookup_obj_attribute(obj_field, rest)
        else:
            return obj_field

    def lookup_field_value(self, context, obj, field):
        """
        Looks up the field value for the passed in object and field name.

        Note that this method is actually called from a template, but this provides a hook
        for subclasses to modify behavior if they wish to do so.

        This may be used for example to change the display value of a variable depending on
        other variables within our context.
        """
        curr_field = field.encode('ascii', 'ignore').decode("utf-8")

        # if this isn't a subfield, check the view to see if it has a get_ method
        if field.find('.') == -1:
            # view supercedes all, does it have a 'get_' method for this obj
            view_method = getattr(self, 'get_%s' % curr_field, None)
            if view_method:
                return view_method(obj)

        return self.lookup_obj_attribute(obj, field)

    def lookup_field_label(self, context, field, default=None):
        """
        Figures out what the field label should be for the passed in field name.

        Our heuristic is as follows:
            1) we check to see if our field_config has a label specified
            2) if not, then we derive a field value from the field name
        """
        # if this is a subfield, strip off everything but the last field name
        if field.find('.') >= 0:
            return self.lookup_field_label(context, field.split('.')[-1], default)

        label = None

        # is there a label specified for this field
        if field in self.field_config and 'label' in self.field_config[field]:
            label = self.field_config[field]['label']

        # if we were given a default, use that
        elif default:
            label = default

        # check our model
        else:
            for model_field in self.model._meta.fields:
                if model_field.name == field:
                    return model_field.verbose_name.title()

        # otherwise, derive it from our field name
        if label is None:
            label = self.derive_field_label(field)

        return label

    def lookup_field_help(self, field, default=None):
        """
        Looks up the help text for the passed in field.
        """
        help = None

        # is there a label specified for this field
        if field in self.field_config and 'help' in self.field_config[field]:
            help = self.field_config[field]['help']

        # if we were given a default, use that
        elif default:
            help = default

        # try to see if there is a description on our model
        elif hasattr(self, 'model'):
            for model_field in self.model._meta.fields:
                if model_field.name == field:
                    help = model_field.help_text
                    break

        return help

    def lookup_field_class(self, field, obj=None, default=None):
        """
        Looks up any additional class we should include when rendering this field
        """
        css = ""

        # is there a class specified for this field
        if field in self.field_config and 'class' in self.field_config[field]:
            css = self.field_config[field]['class']

        # if we were given a default, use that
        elif default:
            css = default

        return css

    def derive_field_label(self, field, obj=None):
        """
        Derives a field label for the passed in field name.
        """
        # replace _'s with ' '
        label = field.replace('_', ' ').title()
        return label

    def derive_field_config(self):
        """
        Derives the field config for this instance.  By default we just use
        self.field_config
        """
        return self.field_config

    def get_template_names(self):
        """
        Returns the name of the template to use to render this request.

        Smartmin provides default templates as fallbacks, so appends it's own templates names to the end
        of whatever list is built by the generic views.

        Subclasses can override this by setting a 'template_name' variable on the class.
        """
        templates = []
        if getattr(self, 'template_name', None):
            templates.append(self.template_name)

        if getattr(self, 'default_template', None):
            templates.append(self.default_template)
        else:
            templates = super(SmartView, self).get_template_names()

        return templates

    def derive_fields(self):
        """
        Default implementation
        """
        fields = []
        if self.fields:
            fields.append(self.fields)

        return fields

    def derive_exclude(self):
        """
        Returns which fields we should exclude
        """
        exclude = []
        if self.exclude:
            exclude += self.exclude

        return exclude

    def derive_refresh(self):
        """
        Returns how many milliseconds before we should refresh
        """
        return self.refresh

    def get_context_data(self, **kwargs):
        """
        We supplement the normal context data by adding our fields and labels.
        """
        context = super(SmartView, self).get_context_data(**kwargs)

        # derive our field config
        self.field_config = self.derive_field_config()

        # add our fields
        self.fields = self.derive_fields()

        # build up our current parameter string, EXCLUSIVE of our page.  These
        # are used to build pagination URLs
        url_params = "?"
        order_params = ""
        for key in self.request.GET.keys():
            if key != 'page' and key != 'pjax' and (len(key) == 0 or key[0] != '_'):
                for value in self.request.GET.getlist(key):
                    url_params += "%s=%s&" % (urlquote(key), urlquote(value))
            elif key == '_order':
                order_params = "&".join(["%s=%s" % (key, _) for _ in self.request.GET.getlist(key)])

        context['url_params'] = url_params
        context['order_params'] = order_params + "&"
        context['pjax'] = self.pjax

        # set our blocks
        context['blocks'] = dict()

        # stuff it all in our context
        context['fields'] = self.fields
        context['view'] = self
        context['field_config'] = self.field_config

        context['title'] = self.derive_title()

        # and any extra context the user specified
        context.update(self.extra_context)

        # by default, our base is 'base.html', but we might be pjax
        base_template = "base.html"
        if 'pjax' in self.request.GET or 'pjax' in self.request.POST:
            base_template = "smartmin/pjax.html"

        if 'HTTP_X_PJAX' in self.request.META:
            base_template = "smartmin/pjax.html"

        context['base_template'] = base_template

        # set our refresh if we have one
        refresh = self.derive_refresh()
        if refresh:
            context['refresh'] = refresh

        return context

    def as_json(self, context):
        """
        Responsible for turning our context into an dict that can then be serialized into an
        JSON response.
        """
        raise NotImplementedError("this view can't be rendered as JSON")

    def render_to_response(self, context, **response_kwargs):
        """
        Overloaded to deal with _format arguments.
        """
        # should we try rendering as JSON?
        if '_format' in self.request.GET and self.request.GET['_format'] == 'json':
            try:
                return JsonResponse(self.as_json(context), safe=False)
            except NotImplementedError:
                pass

        # otherwise, return normally
        return super(SmartView, self).render_to_response(context)


class SmartTemplateView(SmartView, TemplateView):
    pass


def derive_single_object_url_pattern(slug_url_kwarg, path, action):
    """
    Utility function called by class methods for single object views
    """
    if slug_url_kwarg:
        return r'^%s/%s/(?P<%s>[^/]+)/$' % (path, action, slug_url_kwarg)
    else:
        return r'^%s/%s/(?P<pk>\d+)/$' % (path, action)


class SmartSingleObjectView(SmartView):
    slug_field = None
    slug_url_kwarg = None

    def get_slug_field(self):
        """
        If `slug_field` isn't specified it defaults to `slug_url_kwarg`
        """
        return self.slug_field if self.slug_field else self.slug_url_kwarg


class SmartReadView(SmartSingleObjectView, DetailView):
    default_template = 'smartmin/read.html'
    edit_button = None

    field_config = {'modified_blurb': dict(label="Modified"),
                    'created_blurb': dict(label="Created")}

    @classmethod
    def derive_url_pattern(cls, path, action):
        return derive_single_object_url_pattern(cls.slug_url_kwarg, path, action)

    def derive_queryset(self):
        return super(SmartReadView, self).get_queryset()

    def get_queryset(self):
        self.queryset = self.derive_queryset()
        return self.queryset

    def derive_title(self):
        """
        By default we just return the string representation of our object
        """
        return str(self.object)

    def derive_fields(self):
        """
        Derives our fields.  We first default to using our 'fields' variable if available,
        otherwise we figure it out from our object.
        """
        if self.fields:
            return list(self.fields)

        else:
            fields = []
            for field in self.object._meta.fields:
                fields.append(field.name)

            # only exclude?  then remove those items there
            exclude = self.derive_exclude()

            # remove any excluded fields
            fields = [field for field in fields if field not in exclude]

            return fields

    def get_modified_blurb(self, obj):
        return "%s by %s" % (obj.modified_on.strftime("%B %d, %Y at %I:%M %p"), obj.modified_by)

    def get_created_blurb(self, obj):
        return "%s by %s" % (obj.created_on.strftime("%B %d, %Y at %I:%M %p"), obj.created_by)


class SmartDeleteView(SmartSingleObjectView, DetailView, ProcessFormView):
    default_template = 'smartmin/delete_confirm.html'
    name_field = 'name'
    cancel_url = None
    redirect_url = None

    @classmethod
    def derive_url_pattern(cls, path, action):
        return derive_single_object_url_pattern(cls.slug_url_kwarg, path, action)

    def get_cancel_url(self):
        if not self.cancel_url:
            raise ImproperlyConfigured("DeleteView must define a cancel_url")

        return smart_url(self.cancel_url, self.object)

    def pre_delete(self, obj):
        # auto populate modified_by if it is present
        if self.request.user.id and self.request.user.id > 0 and hasattr(obj, 'modified_by_id'):
            obj.modified_by = self.request.user

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.pre_delete(self.object)
        redirect_url = self.get_redirect_url()
        self.object.delete()

        return HttpResponseRedirect(redirect_url)

    def get_redirect_url(self, **kwargs):
        if not self.redirect_url:
            raise ImproperlyConfigured("DeleteView must define a redirect_url")

        return smart_url(self.redirect_url)

    def get_context_data(self, **kwargs):
        """ Add in the field to use for the name field """
        context = super(SmartDeleteView, self).get_context_data(**kwargs)
        context['name_field'] = self.name_field
        context['cancel_url'] = self.get_cancel_url()
        return context


class SmartListView(SmartView, ListView):
    default_template = 'smartmin/list.html'

    link_url = None
    link_fields = None
    add_button = None
    search_fields = None
    paginate_by = 25
    field_config = {'is_active': dict(label='')}
    default_order = None
    select_related = None

    @classmethod
    def derive_url_pattern(cls, path, action):
        if action == 'list':
            return r'^%s/$' % (path)
        else:
            return r'^%s/%s/$' % (path, action)

    def derive_search_fields(self):
        """
        Derives our search fields, by default just returning what was set
        """
        return self.search_fields

    def derive_title(self):
        """
        Derives our title from our list
        """
        title = super(SmartListView, self).derive_title()

        if not title:
            return force_text(self.model._meta.verbose_name_plural).title()
        else:
            return title

    def derive_link_fields(self, context):
        """
        Used to derive which fields should be linked.  This should return a set() containing
        the names of those fields which should be linkable.
        """
        if self.link_fields is not None:
            return self.link_fields

        else:
            link_fields = set()
            if self.fields:
                for field in self.fields:
                    if field != 'is_active':
                        link_fields.add(field)
                        break

        return link_fields

    def lookup_field_link(self, context, field, obj):
        """
        By default we just return /view/{{ id }}/ for the current object.
        """
        return smart_url(self.link_url, obj)

    def lookup_field_orderable(self, field):
        """
        Returns whether the passed in field is sortable or not, by default all 'raw' fields, that
        is fields that are part of the model are sortable.
        """
        try:
            self.model._meta.get_field_by_name(field)
            return True
        except Exception:
            # that field doesn't exist, so not sortable
            return False

    def get_context_data(self, **kwargs):
        """
        Add in what fields are linkable
        """
        context = super(SmartListView, self).get_context_data(**kwargs)

        # our linkable fields
        self.link_fields = self.derive_link_fields(context)

        # stuff it all in our context
        context['link_fields'] = self.link_fields

        # our search term if any
        if 'search' in self.request.GET:
            context['search'] = self.request.GET['search']

        # our ordering field if any
        order = self.derive_ordering()
        if order:
            if order[0] == '-':
                context['order'] = order[1:]
                context['order_asc'] = False
            else:
                context['order'] = order
                context['order_asc'] = True

        return context

    def derive_select_related(self):
        return self.select_related

    def derive_queryset(self, **kwargs):
        """
        Derives our queryset.
        """
        # get our parent queryset
        queryset = super(SmartListView, self).get_queryset(**kwargs)

        # apply any filtering
        search_fields = self.derive_search_fields()
        search_query = self.request.GET.get('search')
        if search_fields and search_query:
            term_queries = []
            for term in search_query.split(' '):
                field_queries = []
                for field in search_fields:
                    field_queries.append(Q(**{field: term}))
                term_queries.append(reduce(operator.or_, field_queries))

            queryset = queryset.filter(reduce(operator.and_, term_queries))

        # add any select related
        related = self.derive_select_related()
        if related:
            queryset = queryset.select_related(*related)

        # return our queryset
        return queryset

    def get_queryset(self, **kwargs):
        """
        Gets our queryset.  This takes care of filtering if there are any
        fields to filter by.
        """
        queryset = self.derive_queryset(**kwargs)

        return self.order_queryset(queryset)

    def derive_ordering(self):
        """
        Returns what field should be used for ordering (using a prepended '-' to indicate descending sort).

        If the default order of the queryset should be used, returns None
        """
        if '_order' in self.request.GET:
            return self.request.GET['_order']
        elif self.default_order:
            return self.default_order
        else:
            return None

    def order_queryset(self, queryset):
        """
        Orders the passed in queryset, returning a new queryset in response.  By default uses the _order query
        parameter.
        """
        order = self.derive_ordering()

        # if we get our order from the request
        # make sure it is a valid field in the list
        if '_order' in self.request.GET:
            if order.lstrip('-') not in self.derive_fields():
                order = None

        if order:
            # if our order is a single string, convert to a simple list
            if isinstance(order, str):
                order = (order,)

            queryset = queryset.order_by(*order)

        return queryset

    def derive_fields(self):
        """
        Derives our fields.
        """
        if self.fields:
            return self.fields

        else:
            fields = []
            for field in self.object_list.model._meta.fields:
                if field.name != 'id':
                    fields.append(field.name)
            return fields

    def get_is_active(self, obj):
        """
        Default implementation of get_is_active which returns a simple div so as to
        render a green dot for active items and nothing for inactive ones.

        Users of SmartModel will get this rendering for free.
        """
        if obj.is_active:
            return '<div class="active_icon"></div>'
        else:
            return ''

    def render_to_response(self, context, **response_kwargs):
        """
        Overloaded to deal with _format arguments.
        """
        # is this a select2 format response?
        if self.request.GET.get('_format', 'html') == 'select2':

            results = []
            for obj in context['object_list']:
                result = None
                if hasattr(obj, 'as_select2'):
                    result = obj.as_select2()

                if not result:
                    result = dict(id=obj.pk, text="%s" % obj)

                results.append(result)

            has_more = context['page_obj'].has_next() if context['page_obj'] else False

            json_data = dict(results=results, err='nil', more=has_more)
            return JsonResponse(json_data)
        # otherwise, return normally
        else:
            return super(SmartListView, self).render_to_response(context)


class SmartCsvView(SmartListView):

    def derive_filename(self):
        filename = getattr(self, 'filename', None)
        if not filename:
            filename = "%s.csv" % self.model._meta.verbose_name.lower()
        return filename

    def render_to_response(self, context, **response_kwargs):
        import csv

        # Create the HttpResponse object with the appropriate CSV header.
        response = HttpResponse(content_type='text/csv; charset=utf-8')
        response['Content-Disposition'] = 'attachment; filename=%s' % self.derive_filename()

        writer = csv.writer(response, quoting=csv.QUOTE_ALL)

        fields = self.derive_fields()

        # build up our header row
        header = []
        for field in fields:
            header.append(str(self.lookup_field_label(dict(), field)))
        writer.writerow([s.encode("utf-8") for s in header])

        # then our actual values
        for obj in self.object_list:
            row = []
            for field in fields:
                row.append(str(self.lookup_field_value(dict(), obj, field)))
            writer.writerow([s.encode("utf-8") for s in row])

        return response


class SmartXlsView(SmartListView):

    def derive_filename(self):
        filename = getattr(self, 'filename', None)
        if not filename:
            filename = "%s.xls" % self.model._meta.verbose_name.lower()
        return filename

    def render_to_response(self, context, **response_kwargs):

        from xlwt import Workbook
        book = Workbook()
        sheet1 = book.add_sheet(self.derive_title())
        fields = self.derive_fields()

        # build up our header row
        for col in range(len(fields)):
            field = fields[col]
            sheet1.write(0, col, str(self.lookup_field_label(dict(), field)))

        # then our actual values
        for row in range(len(self.object_list)):
            obj = self.object_list[row]
            for col in range(len(fields)):
                field = fields[col]
                value = str(self.lookup_field_value(dict(), obj, field))
                # skip the header
                sheet1.write(row + 1, col, value)

        # Create the HttpResponse object with the appropriate header.
        response = HttpResponse(content_type='application/vnd.ms-excel')
        response['Content-Disposition'] = 'attachment; filename=%s' % self.derive_filename()
        book.save(response)
        return response


class SmartFormMixin(object):
    readonly = ()
    field_config = {'modified_blurb': dict(label="Modified"),
                    'created_blurb': dict(label="Created")}
    success_message = None
    submit_button_name = _("Submit")

    def derive_title(self):
        """
        Derives our title from our object
        """
        if not self.title:
            return _("Form")
        else:
            return self.title

    def derive_success_message(self):
        """
        Returns a message to display when this form is successfully saved
        """
        return self.success_message

    def get_form(self):
        """
        Returns an instance of the form to be used in this view.
        """
        self.form = super(SmartFormMixin, self).get_form()

        fields = list(self.derive_fields())

        # apply our field filtering on our form class
        exclude = self.derive_exclude()
        exclude += self.derive_readonly()

        # remove any excluded fields
        for field in exclude:
            if field in self.form.fields:
                del self.form.fields[field]

        if fields is not None:
            # filter out our form fields
            remove = [name for name in self.form.fields.keys() if name not in fields]
            for name in remove:
                del self.form.fields[name]

        # stuff in our referer as the default location for where to return
        location = forms.CharField(widget=forms.widgets.HiddenInput(), required=False)

        if ('HTTP_REFERER' in self.request.META):
            location.initial = self.request.META['HTTP_REFERER']

        # add the location to our form fields
        self.form.fields['loc'] = location

        if fields:
            fields.append('loc')

        # provides a hook to programmatically customize fields before rendering
        for (name, field) in self.form.fields.items():
            field = self.customize_form_field(name, field)
            self.form.fields[name] = field

        return self.form

    def customize_form_field(self, name, field):
        """
        Allows views to customize their form fields.  By default, Smartmin replaces the plain textbox
        date input with it's own DatePicker implementation.
        """
        if isinstance(field, forms.fields.DateField) and isinstance(field.widget, forms.widgets.DateInput):
            field.widget = widgets.DatePickerWidget()
            field.input_formats = [field.widget.input_format[1]] + list(field.input_formats)

        if isinstance(field, forms.fields.ImageField) and isinstance(field.widget, forms.widgets.ClearableFileInput):
            field.widget = widgets.ImageThumbnailWidget()

        return field

    def lookup_field_label(self, context, field, default=None):
        """
        Figures out what the field label should be for the passed in field name.

        We overload this so as to use our form to see if there is label set there.  If so
        then we'll pass that as the default instead of having our parent derive
        the field from the name.
        """
        default = None

        # model forms will have meta inherited from their model that includes labels
        meta_labels = self.form._meta.labels if hasattr(self.form, "_meta") else {}

        if meta_labels and field in meta_labels:
            default = meta_labels[field]
        else:
            for form_field in self.form:
                if form_field.name == field:
                    default = form_field.label
                    break

        return super(SmartFormMixin, self).lookup_field_label(context, field, default=default)

    def lookup_field_help(self, field, default=None):
        """
        Looks up the help text for the passed in field.

        This is overloaded so that we can check whether our form has help text set
        explicitly.  If so, we will pass this as the default to our parent function.
        """
        default = None

        # model forms will have meta inherited from their model that includes help texts
        meta_help_texts = self.form._meta.help_texts if hasattr(self.form, "_meta") else {}

        if meta_help_texts and field in meta_help_texts:
            default = meta_help_texts[field]
        else:
            for form_field in self.form:
                if form_field.name == field:
                    default = form_field.help_text
                    break

        return super(SmartFormMixin, self).lookup_field_help(field, default=default)

    def derive_readonly(self):
        """
        Figures out what fields should be readonly.  We iterate our field_config to find all
        that have a readonly of true
        """
        readonly = list(self.readonly)
        for key, value in self.field_config.items():
            if 'readonly' in value and value['readonly']:
                readonly.append(key)

        return readonly

    def derive_fields(self):
        """
        Derives our fields.
        """
        if self.fields is not None:
            fields = list(self.fields)
        else:
            form = self.form
            fields = []
            for field in form:
                fields.append(field.name)

            # this is slightly confusing but we add in readonly fields here because they will still
            # need to be displayed
            readonly = self.derive_readonly()
            if readonly:
                fields += readonly

        # remove any excluded fields
        for exclude in self.derive_exclude():
            if exclude in fields:
                fields.remove(exclude)

        return fields

    def get_form_class(self):
        """
        Returns the form class to use in this view
        """
        if self.form_class:
            form_class = self.form_class

        else:
            if self.model is not None:
                # If a model has been explicitly provided, use it
                model = self.model
            elif hasattr(self, 'object') and self.object is not None:
                # If this view is operating on a single object, use
                # the class of that object
                model = self.object.__class__
            else:
                # Try to get a queryset and extract the model class
                # from that
                model = self.get_queryset().model

            # run time parameters when building our form
            factory_kwargs = self.get_factory_kwargs()
            form_class = model_forms.modelform_factory(model, **factory_kwargs)

        return form_class

    def get_factory_kwargs(self):
        """
        Let's us specify any extra parameters we might want to call for our form factory.

        These can include: 'form', 'fields', 'exclude' or 'formfield_callback'
        """
        params = dict()

        exclude = self.derive_exclude()
        exclude += self.derive_readonly()

        if self.fields:
            fields = list(self.fields)
            for ex in exclude:
                if ex in fields:
                    fields.remove(ex)

            params['fields'] = fields

        if exclude:
            params['exclude'] = exclude

        return params

    def get_success_url(self):
        """
        By default we use the referer that was stuffed in our
        form when it was created
        """
        if self.success_url:
            # if our smart url references an object, pass that in
            if self.success_url.find('@') > 0:
                return smart_url(self.success_url, self.object)
            else:
                return smart_url(self.success_url, None)

        elif 'loc' in self.form.cleaned_data:
            return self.form.cleaned_data['loc']

        raise ImproperlyConfigured("No redirect location found, override get_success_url to not use redirect urls")

    def derive_initial(self):
        """
        Returns what initial dict should be passed to our form. By default this is empty.
        """
        return dict()

    def get_form_kwargs(self):
        """
        We override this, using only those fields specified if they are specified.

        Otherwise we include all fields in a standard ModelForm.
        """
        kwargs = super(SmartFormMixin, self).get_form_kwargs()
        kwargs['initial'] = self.derive_initial()
        return kwargs

    def derive_submit_button_name(self):
        """
        Returns the name for our button
        """
        return self.submit_button_name

    def get_context_data(self, **kwargs):
        context = super(SmartFormMixin, self).get_context_data(**kwargs)
        context['submit_button_name'] = self.derive_submit_button_name()
        return context


class SmartFormView(SmartFormMixin, SmartView, FormView):
    default_template = 'smartmin/form.html'

    def form_valid(self, form):
        # plug in our success message
        messages.success(self.request, self.derive_success_message())
        return super(SmartFormView, self).form_valid(form)


class SmartModelFormView(SmartFormMixin, SmartSingleObjectView, ModelFormMixin):
    javascript_submit = None

    field_config = {'modified_blurb': dict(label="Modified"), 'created_blurb': dict(label="Created")}

    def derive_title(self):
        """
        Derives our title from our object
        """
        if not self.title:
            return _("Edit %s") % force_text(self.model._meta.verbose_name).title()
        else:
            return self.title

    def pre_save(self, obj):
        """
        Called before an object is saved away
        """
        return obj

    def save(self, obj):
        """
        Actually does the saving of this object, this is when the object is committed
        """
        self.object.save()
        self.save_m2m()

    def form_valid(self, form):
        self.object = form.save(commit=False)

        try:
            self.object = self.pre_save(self.object)
            self.save(self.object)
            self.object = self.post_save(self.object)

            messages.success(self.request, self.derive_success_message())
            if 'HTTP_X_FORMAX' not in self.request.META:
                return HttpResponseRedirect(self.get_success_url())
            else:
                response = self.render_to_response(self.get_context_data(form=form))
                response['REDIRECT'] = self.get_success_url()
                return response

        except IntegrityError as e:
            message = str(e).capitalize()
            errors = self.form._errors.setdefault(forms.forms.NON_FIELD_ERRORS, forms.utils.ErrorList())
            errors.append(message)
            return self.render_to_response(self.get_context_data(form=form))

    def save_m2m(self):
        """
        By default saves the form's m2m, can be overridden if a more complicated m2m model exists
        """
        self.form.save_m2m()

    def post_save(self, obj):
        """
        Called after an object is successfully saved
        """
        return obj

    def get_context_data(self, **kwargs):
        context = super(SmartModelFormView, self).get_context_data(**kwargs)
        context['javascript_submit'] = self.javascript_submit
        return context


class SmartUpdateView(SmartModelFormView, UpdateView):
    default_template = 'smartmin/update.html'
    exclude = ('created_by', 'modified_by')
    submit_button_name = _("Save Changes")

    # allows you to specify the name of URL to use for a remove link that will automatically be shown
    delete_url = None

    @classmethod
    def derive_url_pattern(cls, path, action):
        return derive_single_object_url_pattern(cls.slug_url_kwarg, path, action)

    def derive_queryset(self):
        return super(SmartUpdateView, self).get_queryset()

    def get_queryset(self):
        self.queryset = self.derive_queryset()
        return self.queryset

    def derive_success_message(self):
        # First check whether a default message has been set
        if self.success_message is None:
            return "Your %s has been updated." % self.model._meta.verbose_name
        else:
            return self.success_message

    def pre_save(self, obj):
        # auto populate modified_by if it is present
        if self.request.user.id and self.request.user.id > 0 and hasattr(obj, 'modified_by_id'):
            obj.modified_by = self.request.user

        return obj

    def get_context_data(self, **kwargs):
        context = super(SmartUpdateView, self).get_context_data(**kwargs)

        if self.delete_url:
            context['delete_url'] = smart_url(self.delete_url, self.object)

        return context

    def get_modified_blurb(self, obj):
        return "%s by %s" % (obj.modified_on.strftime("%B %d, %Y at %I:%M %p"), obj.modified_by)

    def get_created_blurb(self, obj):
        return "%s by %s" % (obj.created_on.strftime("%B %d, %Y at %I:%M %p"), obj.created_by)


class SmartModelActionView(SmartFormMixin, SmartSingleObjectView, DetailView, ProcessFormView):

    @classmethod
    def derive_url_pattern(cls, path, action):
        return derive_single_object_url_pattern(cls.slug_url_kwarg, path, action)

    def execute_action(self):
        """
        Subclasses should do their work here. They can throw a ValidationError to return
        control back to the Form input with said error. If no error is thrown page will
        be redirected to success_url
        """
        pass

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(SmartModelActionView, self).post(request, *args, **kwargs)

    def form_valid(self, form):
        try:
            self.execute_action()

        except forms.ValidationError as e:
            # turns out we aren't valid after all, stuff our error into our form
            self.form.add_error(None, e)
            return self.form_invalid(form)

        # all went well, stuff our success message in and return
        messages.success(self.request, self.derive_success_message())
        return super(SmartModelActionView, self).form_valid(form)


class SmartMultiFormView(SmartView, TemplateView):
    default_template = 'smartmin/multi_form.html'
    forms = {}

    # allows you to specify the name of URL to use for a remove link that will automatically be shown
    delete_url = None

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        page_forms = []
        for prefix, form in self.forms.items():
            f = form(prefix=prefix)
            page_forms.append(f)

        context['forms'] = page_forms

        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        # process our forms
        page_forms = []
        valid = True
        for prefix, form in self.forms.items():
            f = form(request.POST, prefix=prefix)
            valid = valid and f.is_valid()
            page_forms.append(f)

        if not valid:
            context['forms'] = page_forms
            return self.render_to_response(context)
        else:
            # redirect to success page
            pass

    def get_context_data(self, **kwargs):
        context = super(SmartMultiFormView, self).get_context_data(**kwargs)

        if self.delete_url:
            context['delete_url'] = smart_url(self.delete_url, self.object)

        return context


class SmartCreateView(SmartModelFormView, CreateView):
    default_template = 'smartmin/create.html'
    exclude = ('created_by', 'modified_by', 'is_active')
    submit_button_name = _("Create")

    def has_object_permission(self, getter_name):
        # create views don't have an object, so this is always False
        return False

    def pre_save(self, obj):
        if self.request.user.id and self.request.user.id > 0:
            # auto populate created_by if it is present
            if hasattr(obj, 'created_by_id'):
                obj.created_by = self.request.user

            # auto populate modified_by if it is present
            if hasattr(obj, 'modified_by_id'):
                obj.modified_by = self.request.user

        return obj

    def derive_success_message(self):
        # First check whether a default message has been set
        if self.success_message is None:
            return _("Your new %s has been created.") % self.model._meta.verbose_name
        else:
            return self.success_message

    def derive_title(self):
        """
        Derives our title from our object
        """
        if not self.title:
            return _("Create %s") % force_text(self.model._meta.verbose_name).title()
        else:
            return self.title


class SmartCSVImportView(NonAtomicMixin, SmartCreateView):
    success_url = 'id@csv_imports.importtask_read'

    fields = ('csv_file',)

    def derive_title(self):
        return _("Import %s") % self.crudl.model._meta.verbose_name_plural.title()

    def pre_save(self, obj):
        obj = super(SmartCSVImportView, self).pre_save(obj)
        obj.model_class = "%s.%s" % (self.crudl.model.__module__, self.crudl.model.__name__)
        return obj

    def post_save(self, task):
        task = super(SmartCSVImportView, self).post_save(task)

        task.import_params = json.dumps(self.form.data)

        # kick off our CSV import
        task.start()

        return task


class SmartCRUDL(object):
    actions = ('create', 'read', 'update', 'delete', 'list')
    model_name = None
    app_name = None
    module_name = None
    path = None

    permissions = True

    def __init__(self, model=None, path=None, actions=None):
        # set our model if passed in
        if model:
            self.model = model

        # derive our model name
        if not self.model_name:
            self.model_name = self.model._meta.object_name

        # derive our app name
        if not self.app_name:
            self.app_name = self.model._meta.app_label

        # derive our path from our class name
        if not path and not self.path:
            self.path = self.model_name.lower()

        # derive our module name from our class's module
        if not self.module_name:
            parts = self.__class__.__module__.split(".")
            self.module_name = parts[-2]

            # deal with special case of views subdirectories, we need to go up one more to find the real module
            if self.module_name == 'views' and len(parts) >= 3:
                self.module_name = parts[-3]

        # set our actions if set
        if actions:
            self.actions = actions

    def permission_for_action(self, action):
        """
        Returns the permission to use for the passed in action
        """
        return "%s.%s_%s" % (self.app_name.lower(), self.model_name.lower(), action)

    def template_for_action(self, action):
        """
        Returns the template to use for the passed in action
        """
        return "%s/%s_%s.html" % (self.module_name.lower(), self.model_name.lower(), action)

    def url_name_for_action(self, action):
        """
        Returns the reverse name for this action
        """
        return "%s.%s_%s" % (self.module_name.lower(), self.model_name.lower(), action)

    def view_for_action(self, action):
        """
        Returns the appropriate view class for the passed in action
        """
        # this turns replace_foo into ReplaceFoo and read into Read
        class_name = "".join([word.capitalize() for word in action.split("_")])
        view = None

        # see if we have a custom class defined for this action
        if hasattr(self, class_name):
            # return that one
            view = getattr(self, class_name)

            # no model set?  set it ourselves
            if not getattr(view, 'model', None):
                view.model = self.model

            # no permission and we are supposed to set them, do so
            if not hasattr(view, 'permission') and self.permissions:
                view.permission = self.permission_for_action(action)

            # set our link URL based on read and update
            if not getattr(view, 'link_url', None):
                if 'read' in self.actions:
                    view.link_url = 'id@%s' % self.url_name_for_action('read')
                elif 'update' in self.actions:
                    view.link_url = 'id@%s' % self.url_name_for_action('update')

            # if we can't infer a link URL then view class must override lookup_field_link
            if not getattr(view, 'link_url', None) and 'lookup_field_link' not in view.__dict__:
                view.link_fields = ()

            # set add_button based on existence of Create view if add_button not explicitly set
            if action == 'list' and getattr(view, 'add_button', None) is None:
                view.add_button = 'create' in self.actions

            # set edit_button based on existence of Update view if edit_button not explicitly set
            if action == 'read' and getattr(view, 'edit_button', None) is None:
                view.edit_button = 'update' in self.actions

            # if update or create, set success url if not set
            if not getattr(view, 'success_url', None) and (action == 'update' or action == 'create'):
                view.success_url = '@%s' % self.url_name_for_action('list')

        # otherwise, use our defaults
        else:
            options = dict(model=self.model)

            # if this is an update or create, and we have a list view, then set the default to that
            if action == 'update' or action == 'create' and 'list' in self.actions:
                options['success_url'] = '@%s' % self.url_name_for_action('list')

            # set permissions if appropriate
            if self.permissions:
                options['permission'] = self.permission_for_action(action)

            if action == 'create':
                view = type(str("%sCreateView" % self.model_name), (SmartCreateView,), options)

            elif action == 'read':
                if 'update' in self.actions:
                    options['edit_button'] = True

                view = type(str("%sReadView" % self.model_name), (SmartReadView,), options)

            elif action == 'update':
                if 'delete' in self.actions:
                    options['delete_url'] = 'id@%s' % self.url_name_for_action('delete')

                view = type(str("%sUpdateView" % self.model_name), (SmartUpdateView,), options)

            elif action == 'delete':
                if 'list' in self.actions:
                    options['cancel_url'] = '@%s' % self.url_name_for_action('list')
                    options['redirect_url'] = '@%s' % self.url_name_for_action('list')

                elif 'update' in self.actions:
                    options['cancel_url'] = '@%s' % self.url_name_for_action('update')

                view = type(str("%sDeleteView" % self.model_name), (SmartDeleteView,), options)

            elif action == 'list':
                if 'read' in self.actions:
                    options['link_url'] = 'id@%s' % self.url_name_for_action('read')
                elif 'update' in self.actions:
                    options['link_url'] = 'id@%s' % self.url_name_for_action('update')
                else:
                    options['link_fields'] = ()

                if 'create' in self.actions:
                    options['add_button'] = True

                view = type(str("%sListView" % self.model_name), (SmartListView,), options)

            elif action == 'csv_import':
                options['model'] = ImportTask
                view = type(str("%sCSVImportView" % self.model_name), (SmartCSVImportView,), options)

        if not view:
            # couldn't find a view?  blow up
            raise Exception("No view found for action: %s" % action)

        # set the url name for this view
        view.url_name = self.url_name_for_action(action)

        # no template set for it?  set one based on our action and app name
        if not getattr(view, 'template_name', None):
            view.template_name = self.template_for_action(action)

        view.crudl = self

        return view

    def pattern_for_view(self, view, action):
        """
        Returns the URL pattern for the passed in action.
        """
        # if this view knows how to define a URL pattern, call that
        if getattr(view, 'derive_url_pattern', None):
            return view.derive_url_pattern(self.path, action)

        # otherwise take our best guess
        else:
            return r'^%s/%s/$' % (self.path, action)

    def as_urlpatterns(self):
        """
        Creates the appropriate URLs for this object.
        """
        urls = []

        # for each of our actions
        for action in self.actions:
            view_class = self.view_for_action(action)
            view_pattern = self.pattern_for_view(view_class, action)
            name = self.url_name_for_action(action)
            urls.append(url(view_pattern, view_class.as_view(), name=name))

        return urls



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csv_imports', '0002_auto_20161118_1920'),
    ]

    operations = [
        migrations.AddField(
            model_name='importtask',
            name='task_status',
            field=models.CharField(default='PENDING', max_length=32),
        ),
    ]



from unittest.mock import patch, call

from django.core.management import call_command
from django.db.migrations import RunSQL, RunPython
from django.test import TestCase

from .commands.collect_sql import SqlType, SqlObjectOperation


class MockMigration(object):
    def __init__(self, operations):
        self.operations = operations


def mock_run_python(apps, schema_editor):
    pass


class CollectSqlTest(TestCase):

    @patch('smartmin.management.commands.collect_sql.Command.load_migrations')
    @patch('smartmin.management.commands.collect_sql.Command.write_dump')
    def test_command(self, mock_write_dump, mock_load_migrations):
        mock_load_migrations.return_value = [
            MockMigration(operations=[
                RunSQL("""
CREATE INDEX test_1 ON foo(bar);
CREATE INDEX test_2 ON foo(bar); create unique index test_3 on foo(bar);
"""),
                RunPython(mock_run_python)
            ]),
            MockMigration(operations=[
                RunSQL("DROP INDEX test_2;"),
            ]),
            MockMigration(operations=[
                RunSQL("CREATE TRIGGER test_1 AFTER TRUNCATE ON flows_flowstep EXECUTE PROCEDURE foo();"),
                RunSQL("CREATE INDEX a_test ON foo(bar);"),
                RunPython(mock_run_python)
            ]),
            MockMigration(operations=[
                RunSQL("CREATE OR REPLACE FUNCTION a_func(i INTEGER, s TEXT) RETURNS integer AS $$ BEGIN RETURN i + 1; END; $$ LANGUAGE plpgsql;"),

                # different function because it has different parameters
                RunSQL("CREATE OR REPLACE FUNCTION a_func(s text, integer) RETURNS text AS $$ BEGIN RETURN UPPER(s); END; $$ LANGUAGE plpgsql;"),

                # same function because it has same name and parameters
                RunSQL("CREATE OR REPLACE FUNCTION A_FUNC(N integer, V text) RETURNS text AS $$ BEGIN RETURN UPPER(V); END; $$ LANGUAGE plpgsql;"),

                RunSQL("CREATE OR REPLACE FUNCTION func2(i INTEGER) RETURNS integer AS $$ BEGIN RETURN i + 1; END; $$ LANGUAGE plpgsql;"),
                RunSQL("DROP FUNCTION func2(i INTEGER);"),
            ])
        ]

        call_command('collect_sql', output_dir='sql')

        mock_write_dump.assert_has_calls([
            call('indexes', [
                SqlObjectOperation("CREATE INDEX a_test ON foo(bar);", sql_type=SqlType.INDEX, obj_name="a_test", is_create=True),
                SqlObjectOperation("CREATE INDEX test_1 ON foo(bar);", sql_type=SqlType.INDEX, obj_name="test_1", is_create=True),
                SqlObjectOperation("create unique index test_3 on foo(bar);", sql_type=SqlType.INDEX, obj_name="test_3", is_create=True),
            ], 'sql'),
            call('functions', [
                SqlObjectOperation(
                    "CREATE OR REPLACE FUNCTION A_FUNC(N integer, V text) RETURNS text AS $$ BEGIN RETURN UPPER(V); END; $$ LANGUAGE plpgsql;",
                    sql_type=SqlType.FUNCTION,
                    obj_name="A_FUNC(integer,text)",
                    is_create=True,
                ),
                SqlObjectOperation(
                    "CREATE OR REPLACE FUNCTION a_func(s text, integer) RETURNS text AS $$ BEGIN RETURN UPPER(s); END; $$ LANGUAGE plpgsql;",
                    sql_type=SqlType.FUNCTION,
                    obj_name="a_func(text,integer)",
                    is_create=True,
                ),
            ], 'sql'),
            call('triggers', [
                SqlObjectOperation(
                    "CREATE TRIGGER test_1 AFTER TRUNCATE ON flows_flowstep EXECUTE PROCEDURE foo();",
                    sql_type=SqlType.TRIGGER,
                    obj_name="test_1",
                    is_create=True,
                ),
            ], 'sql')
        ])



import sys

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import Permission, Group
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_migrate

from smartmin.perms import assign_perm, remove_perm

permissions_app_name = None


def get_permissions_app_name():
    """
    Gets the app after which smartmin permissions should be installed. This can be specified by PERMISSIONS_APP in the
    Django settings or defaults to the last app with models
    """
    global permissions_app_name

    if not permissions_app_name:
        permissions_app_name = getattr(settings, 'PERMISSIONS_APP', None)

        if not permissions_app_name:
            app_names_with_models = [a.name for a in apps.get_app_configs() if a.models_module is not None]
            if app_names_with_models:
                permissions_app_name = app_names_with_models[-1]

    return permissions_app_name


def is_permissions_app(app_config):
    """
    Returns whether this is the app after which permissions should be installed.
    """
    return app_config.name == get_permissions_app_name()


def check_role_permissions(role, permissions, current_permissions):
    """
    Checks the the passed in role (can be user, group or AnonymousUser)  has all the passed
    in permissions, granting them if necessary.
    """
    role_permissions = []

    # get all the current permissions, we'll remove these as we verify they should still be granted
    for permission in permissions:
        splits = permission.split(".")
        if len(splits) != 2 and len(splits) != 3:
            sys.stderr.write("  invalid permission %s, ignoring\n" % permission)
            continue

        app = splits[0]
        codenames = []

        if len(splits) == 2:
            codenames.append(splits[1])
        else:
            (object, action) = splits[1:]

            # if this is a wildcard, then query our database for all the permissions that exist on this object
            if action == '*':
                for perm in Permission.objects.filter(codename__startswith="%s_" % object, content_type__app_label=app):
                    codenames.append(perm.codename)
            # otherwise, this is an error, continue
            else:
                sys.stderr.write("  invalid permission %s, ignoring\n" % permission)
                continue

        if len(codenames) == 0:
            continue

        for codename in codenames:
            # the full codename for this permission
            full_codename = "%s.%s" % (app, codename)

            # this marks all the permissions which should remain
            role_permissions.append(full_codename)

            try:
                assign_perm(full_codename, role)
            except ObjectDoesNotExist:
                pass
                # sys.stderr.write("  unknown permission %s, ignoring\n" % permission)

    # remove any that are extra
    for permission in current_permissions:
        if isinstance(permission, str):
            key = permission
        else:
            key = "%s.%s" % (permission.content_type.app_label, permission.codename)

        if key not in role_permissions:
            remove_perm(key, role)


def check_all_group_permissions(sender, **kwargs):
    """
    Checks that all the permissions specified in our settings.py are set for our groups.
    """
    if not is_permissions_app(sender):
        return

    config = getattr(settings, 'GROUP_PERMISSIONS', dict())

    # for each of our items
    for name, permissions in config.items():
        # get or create the group
        (group, created) = Group.objects.get_or_create(name=name)
        if created:
            pass

        check_role_permissions(group, permissions, group.permissions.all())


def add_permission(content_type, permission):
    """
    Adds the passed in permission to that content type.  Note that the permission passed
    in should be a single word, or verb.  The proper 'codename' will be generated from that.
    """
    # build our permission slug
    codename = "%s_%s" % (content_type.model, permission)

    # sys.stderr.write("Checking %s permission for %s\n" % (permission, content_type.name))

    # does it already exist
    if not Permission.objects.filter(content_type=content_type, codename=codename):
        Permission.objects.create(content_type=content_type,
                                  codename=codename,
                                  name="Can %s %s" % (permission, content_type.name))
        # sys.stderr.write("Added %s permission for %s\n" % (permission, content_type.name))


def check_all_permissions(sender, **kwargs):
    """
    This syncdb checks our PERMISSIONS setting in settings.py and makes sure all those permissions
    actually exit.
    """
    if not is_permissions_app(sender):
        return

    config = getattr(settings, 'PERMISSIONS', dict())

    # for each of our items
    for natural_key, permissions in config.items():
        # if the natural key '*' then that means add to all objects
        if natural_key == '*':
            # for each of our content types
            for content_type in ContentType.objects.all():
                for permission in permissions:
                    add_permission(content_type, permission)

        # otherwise, this is on a specific content type, add for each of those
        else:
            app, model = natural_key.split('.')
            try:
                content_type = ContentType.objects.get_by_natural_key(app, model)
            except ContentType.DoesNotExist:
                continue

            # add each permission
            for permission in permissions:
                add_permission(content_type, permission)


post_migrate.connect(check_all_permissions)
post_migrate.connect(check_all_group_permissions)



"""
simple mixins that keep you from writing so much code
"""

from django.db import transaction
from django.utils.decorators import method_decorator


class PassRequestToFormMixin(object):
    """
    Mixin to include the request in the form kwargs
    """
    def get_form_kwargs(self):
        kwargs = super(PassRequestToFormMixin, self).get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs


class NonAtomicMixin(object):
    """
    Mixin to configure a view to be handled without a transaction
    """
    @method_decorator(transaction.non_atomic_requests)
    def dispatch(self, request, *args, **kwargs):
        return super(NonAtomicMixin, self).dispatch(request, *args, **kwargs)



import os
import sys
from io import StringIO

from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponse
from django.utils import timezone


class AjaxRedirect:

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        if request.is_ajax():
            if type(response) == HttpResponseRedirect:
                # This is our own AJAX friend redirect to allow
                # the calling Javascript the opportunity to deal
                # with redirect responses in its own way
                response = HttpResponse(response["Location"])
                response.status_code = 302
        return response


class ProfileMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, view, *args, **kwargs):
        import hotshot, hotshot.stats  # noqa

        for item in request.META['QUERY_STRING'].split('&'):
            if item.split('=')[0] == 'profile':  # profile in query string

                # catch the output, must happen before stats object is created
                # see https://bugs.launchpad.net/webpy/+bug/133080 for the details
                std_old, std_new = sys.stdout, StringIO.StringIO()
                sys.stdout = std_new

                # now let's do some profiling
                tmpfile = '/tmp/%s' % request.COOKIES['sessionid']
                prof = hotshot.Profile(tmpfile)

                # make a call to the actual view function with the given arguments
                prof.runcall(view, request, *args[0], **args[1])
                prof.close()

                # and then statistical reporting
                stats = hotshot.stats.load(tmpfile)
                stats.strip_dirs()
                stats.sort_stats('time')

                # do the output
                stats.print_stats(1.0)

                # restore default output
                sys.stdout = std_old

                # delete file
                os.remove(tmpfile)

                return HttpResponse('<pre>%s</pre>' % std_new.getvalue())

        return None


class TimezoneMiddleware:

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        user_tz = getattr(settings, 'USER_TIME_ZONE', None)

        if user_tz:
            timezone.activate(user_tz)

        return response



from django.db import migrations, models
import smartmin.csv_imports.models


class Migration(migrations.Migration):

    dependencies = [
        ('csv_imports', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='importtask',
            name='csv_file',
            field=models.FileField(help_text='A comma delimited file of records to import', upload_to=smartmin.csv_imports.models.generate_file_path, verbose_name='Import file'),
        ),
    ]



import os
from io import StringIO

import ho.pisa as pisa

from django.conf import settings
from django.http import HttpResponse
from django.utils.html import escape


class PDFMixin(object):
    """
    Mixin that will change a class based view to render as PDF

    Dependencies:
     - reportlab
     - html5lib
     - pisa
    """
    def render_to_response(self, context, **response_kwargs):

        response = super(PDFMixin, self).render_to_response(context, **response_kwargs)

        # do the actual rendering
        response.render()

        # and get the content
        result = StringIO.StringIO()

        # now render with pisa as PDF
        pdf = pisa.pisaDocument(StringIO.StringIO(response.rendered_content.encode("ISO-8859-1")), result,
                                link_callback=fetch_resource)
        if not pdf.err:
            return HttpResponse(result.getvalue(), mimetype='application/pdf')
        return HttpResponse('We had some errors<pre>%s</pre>' % escape(response.content))


def fetch_resource(uri, rel):
    path = os.path.join(settings.STATICFILES_DIRS[0], uri.replace(settings.STATIC_URL, ""))
    return path



from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('csv_imports', '0003_importtask_task_status'),
    ]

    operations = [
        migrations.AlterField(
            model_name='importtask',
            name='created_on',
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now, editable=False, help_text='When this item was originally created'),
        ),
        migrations.AlterField(
            model_name='importtask',
            name='modified_on',
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now, editable=False, help_text='When this item was last modified'),
        ),
    ]



import django.views.static
from django.conf import settings
from django.urls import reverse
from django.http import HttpResponseRedirect

from .models import PasswordHistory


class ChangePasswordMiddleware:
    """
    Redirects all users to the password change form if we find that a user's
    password is expired.
    """
    def __init__(self, get_response=None):
        self.get_response = get_response

        self.password_expire = getattr(settings, 'USER_PASSWORD_EXPIRATION', -1)

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, view, *args, **kwargs):
        newpassword_path = reverse('users.user_newpassword', args=[0])
        logout_path = reverse('users.user_logout')

        if (self.password_expire < 0 or not request.user.is_authenticated or view == django.views.static.serve or request.path == newpassword_path or request.path == logout_path):  # noqa
            return

        if PasswordHistory.is_password_expired(request.user):
            return HttpResponseRedirect(reverse('users.user_newpassword', args=['0']))



from django.contrib.auth.models import Permission


def assign_perm(perm, group):
    """
    Assigns a permission to a group
    """
    if not isinstance(perm, Permission):
        try:
            app_label, codename = perm.split('.', 1)
        except ValueError:
            raise ValueError("For global permissions, first argument must be in"
                             " format: 'app_label.codename' (is %r)" % perm)
        perm = Permission.objects.get(content_type__app_label=app_label, codename=codename)

    group.permissions.add(perm)
    return perm


def remove_perm(perm, group):
    """
    Removes a permission from a group
    """
    if not isinstance(perm, Permission):
        try:
            app_label, codename = perm.split('.', 1)
        except ValueError:
            raise ValueError("For global permissions, first argument must be in"
                             " format: 'app_label.codename' (is %r)" % perm)
        perm = Permission.objects.get(content_type__app_label=app_label, codename=codename)

    group.permissions.remove(perm)
    return




from collections import OrderedDict
from datetime import datetime
from enum import Enum

import sqlparse
from sqlparse import sql, tokens
from sqlparse import tokens as sql_tokens
from textwrap import dedent

from django.core.management.base import BaseCommand
from django.db.migrations import RunSQL
from django.db.migrations.executor import MigrationExecutor


class InvalidSQLException(Exception):
    def __init__(self, s):
        super(InvalidSQLException, self).__init__(f"Invalid SQL: {s}")


class SqlType(Enum):
    """
    The different SQL types that we can extract from migrations
    """
    INDEX = 1
    FUNCTION = 2
    TRIGGER = 3


class SqlObjectOperation:
    def __init__(self, statement, sql_type, obj_name, is_create):
        self.statement = statement
        self.sql_type = sql_type
        self.obj_name = obj_name
        self.is_create = is_create

    @classmethod
    def parse(cls, raw):
        # get non-whitespace non-comment tokens
        tokens = [t for t in raw.tokens if not t.is_whitespace and not isinstance(t, sql.Comment)]
        if len(tokens) < 3:
            return None

        # check statement is of form "CREATE|DROP TYPE ..."
        if tokens[0].ttype != sql_tokens.DDL or tokens[1].ttype != sql_tokens.Keyword:
            return None

        if tokens[0].value.upper() in ('CREATE', 'CREATE OR REPLACE'):
            is_create = True
        elif tokens[0].value.upper() in ('DROP',):
            is_create = False
        else:
            return None

        try:
            type_token = tokens[1].value.upper()
            if type_token == "UNIQUE":
                type_token = tokens[2].value.upper()

            sql_type = SqlType[type_token]
        except KeyError:
            return None

        if sql_type == SqlType.FUNCTION:
            function = next(filter(lambda t: isinstance(t, sql.Function), tokens), None)
            if not function:
                raise InvalidSQLException(raw.value)

            name = cls._function_name(function)
        else:
            identifier = next(filter(lambda t: isinstance(t, sql.Identifier), tokens), None)
            if not identifier:
                raise InvalidSQLException(raw.value)

            name = identifier.value

        return cls(raw.value.strip(), sql_type, name, is_create)

    @staticmethod
    def _function_name(func: sql.Function) -> str:
        # possible bug in sqlparse library but Function.get_parameters only returns first identifier
        parenthesis = func.tokens[-1]
        types = [str(t) for t in parenthesis.tokens if t.ttype == tokens.Name.Builtin]

        return f"{func.get_name()}({','.join(types)})"

    def __eq__(self, other):
        return self.statement == other.statement and self.sql_type == other.sql_type \
               and self.obj_name == other.obj_name and self.is_create == other.is_create

    def __repr__(self):
        return f"SqlObjectOperation[type={self.sql_type} name='{self.obj_name}' is_create={self.is_create}]"


class Command(BaseCommand):
    help = """Collects SQL statements from migrations to compile lists of indexes, functions and triggers"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--preserve-order', action='store_true', dest='preserve_order', default=False,
            help='Whether to preserve order of operations rather than sorting by object name.',
        )
        parser.add_argument(
            '--output-dir', action='store', dest='output_dir', default='temba/sql',
            help='The output directory for generated SQL files.',
        )

    def handle(self, *args, **options):
        preserve_order = options.get('preserve_order')
        output_dir = options.get('output_dir')

        self.verbosity = options.get('verbosity')

        self.stdout.write("Loading migrations...")

        migrations = self.load_migrations()

        self.stdout.write("Loaded %s migrations" % self.style.SUCCESS(len(migrations)))
        self.stdout.write("Extracting SQL operations...")

        operations = self.extract_operations(migrations)

        self.stdout.write("Extracted %s SQL operations" % self.style.SUCCESS(len(operations)))
        self.stdout.write("Normalizing SQL operations...")

        normalized = self.normalize_operations(operations)

        self.stdout.write("Removed %s redundant operations" % self.style.SUCCESS(len(operations) - len(normalized)))

        self.write_type_dumps(normalized, preserve_order, output_dir)

    def load_migrations(self):  # pragma: no cover
        """
        Loads all migrations in the order they would be applied to a clean database
        """
        executor = MigrationExecutor(connection=None)

        # create the forwards plan Django would follow on an empty database
        plan = executor.migration_plan(executor.loader.graph.leaf_nodes(), clean_start=True)

        if self.verbosity >= 2:
            for migration, _ in plan:
                self.stdout.write(" > %s" % migration)

        return [m[0] for m in plan]

    def extract_operations(self, migrations):
        """
        Extract SQL operations from the given migrations
        """
        operations = []

        for migration in migrations:
            for operation in migration.operations:
                if isinstance(operation, RunSQL):
                    statements = sqlparse.parse(dedent(operation.sql))

                    for statement in statements:
                        operation = SqlObjectOperation.parse(statement)
                        if operation:
                            operations.append(operation)

                            if self.verbosity >= 2:
                                self.stdout.write(" > % -100s (%s)" % (operation, migration))

        return operations

    def normalize_operations(self, operations):
        """
        Removes redundant SQL operations - e.g. a CREATE X followed by a DROP X
        """
        normalized = OrderedDict()

        for operation in operations:
            op_key = (operation.sql_type, operation.obj_name.lower())

            # do we already have an operation for this object?
            if op_key in normalized:
                if self.verbosity >= 2:
                    self.stdout.write(" < %s" % normalized[op_key])

                del normalized[op_key]

            # don't add DROP operations for objects not previously created
            if operation.is_create:
                normalized[op_key] = operation
            elif self.verbosity >= 2:
                self.stdout.write(" < %s" % operation)

        return normalized.values()

    def write_type_dumps(self, operations, preserve_order, output_dir):
        """
        Splits the list of SQL operations by type and dumps these to separate files
        """
        by_type = {SqlType.INDEX: [], SqlType.FUNCTION: [], SqlType.TRIGGER: []}
        for operation in operations:
            by_type[operation.sql_type].append(operation)

        # optionally sort each operation list by the object name
        if not preserve_order:
            for obj_type, ops in by_type.items():
                by_type[obj_type] = sorted(ops, key=lambda o: o.obj_name)

        if by_type[SqlType.INDEX]:
            self.write_dump('indexes', by_type[SqlType.INDEX], output_dir)
        if by_type[SqlType.FUNCTION]:
            self.write_dump('functions', by_type[SqlType.FUNCTION], output_dir)
        if by_type[SqlType.TRIGGER]:
            self.write_dump('triggers', by_type[SqlType.TRIGGER], output_dir)

    def write_dump(self, type_label, operations, output_dir):  # pragma: no cover
        filename = '%s/current_%s.sql' % (output_dir, type_label)

        with open(filename, 'w') as f:
            header = '-- Generated by collect_sql on %s UTC\n\n' % datetime.utcnow().strftime("%Y-%m-%d %H:%M")
            f.write(header)

            for operation in operations:
                f.write(operation.statement)
                f.write('\n\n')

        self.stdout.write("Saved %s" % filename)



# Generated by Django 2.2.17 on 2021-02-19 14:03

from django.db import migrations


def delete_failed_logins(apps, schema_editor):  # pragma: no cover
    FailedLogin = apps.get_model("users", "FailedLogin")

    FailedLogin.objects.all().delete()


def noop(apps, schema_editor):  # pragma: no cover
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(delete_failed_logins, noop)

    ]



from io import StringIO

from celery import shared_task

from django.utils import timezone
from django.utils.module_loading import import_string

from .models import ImportTask


@shared_task(track_started=True)
def csv_import(task_id):  # pragma: no cover
    task_obj = ImportTask.objects.get(pk=task_id)
    log = StringIO()

    task_obj.task_id = csv_import.request.id
    task_obj.task_status = ImportTask.RUNNING
    task_obj.log("Started import at %s" % timezone.now())
    task_obj.log("--------------------------------")
    task_obj.save()

    try:
        model = import_string(task_obj.model_class)
        records = model.import_csv(task_obj, log)
        task_obj.task_status = ImportTask.SUCCESS
        task_obj.save()

        task_obj.log(log.getvalue())
        task_obj.log("Import finished at %s" % timezone.now())
        task_obj.log("%d record(s) added." % len(records))

    except Exception as e:
        import traceback
        traceback.print_exc()

        task_obj.task_status = ImportTask.FAILURE

        task_obj.log("\nError: %s\n" % str(e))
        task_obj.log(log.getvalue())
        task_obj.save()

        raise e

    # give our model the opportunity to do any last finalization outside our transaction
    model.finalize_import(task_obj, records)

    return task_obj



import csv
import datetime
import traceback
import json
import pytz

from django.conf import settings
from django.db import models
from django.utils import timezone

from xlrd import open_workbook, xldate_as_tuple, XL_CELL_DATE, XLRDError


class SmartImportRowError(Exception):
    def __init__(self, message):
        self.message = message

    def __unicode__(self):  # pragma: no cover
        return self.message

    def __str__(self):
        return str(self.__unicode__())


class SmartModel(models.Model):
    """
    Useful abstract base class that adds the concept of something being active,
    having a user that created or modified the item and creation and modification
    dates.
    """
    is_active = models.BooleanField(default=True,
                                    help_text="Whether this item is active, use this instead of deleting")

    created_by = models.ForeignKey(settings.AUTH_USER_MODEL,
                                   on_delete=models.PROTECT,
                                   related_name="%(app_label)s_%(class)s_creations",
                                   help_text="The user which originally created this item")
    created_on = models.DateTimeField(default=timezone.now, editable=False, blank=True,
                                      help_text="When this item was originally created")

    modified_by = models.ForeignKey(settings.AUTH_USER_MODEL,
                                    on_delete=models.PROTECT,
                                    related_name="%(app_label)s_%(class)s_modifications",
                                    help_text="The user which last modified this item")
    modified_on = models.DateTimeField(default=timezone.now, editable=False, blank=True,
                                       help_text="When this item was last modified")

    def save(self, *args, **kwargs):
        update_fields = kwargs.get('update_fields', None)

        if (update_fields is None or 'modified_on' in update_fields) and not kwargs.pop('preserve_modified_on', False):
            self.modified_on = timezone.now()

        return super(SmartModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True

    @classmethod
    def prepare_fields(cls, field_dict, import_params=None, user=None):
        return field_dict

    @classmethod
    def create_instance(cls, field_dict):
        return cls.objects.create(**field_dict)

    @classmethod
    def validate_import_header(cls, header):
        return

    @classmethod
    def get_import_file_headers(cls, csv_file):
        filename = csv_file
        headers = []
        try:
            workbook = open_workbook(filename.name, 'rb')

            for sheet in workbook.sheets():

                # read our header
                header = []
                for col in range(sheet.ncols):
                    header.append(str(sheet.cell(0, col).value))
                headers = [cls.normalize_value(_).lower() for _ in header]

                # only care for the first sheet
                break
        except XLRDError:
            # our alternative codec, by default we are the crazy windows encoding
            ascii_codec = 'cp1252'

            # read the entire file, look for mac_roman characters
            reader = open(filename.name, "rb")
            for byte in reader.read():
                # these are latin accented characterse in mac_roman, if we see them then our alternative
                # encoding should be mac_roman
                try:
                    byte_number = ord(byte)
                except TypeError:
                    byte_number = byte

                if byte_number in [0x81, 0x8d, 0x8f, 0x90, 0x9d]:
                    ascii_codec = 'mac_roman'
                    break
            reader.close()

            reader = open(filename.name, "rU", encoding='utf-8-sig')

            def unicode_csv_reader(utf8_data, dialect=csv.excel, **kwargs):
                csv_reader = csv.reader(utf8_data, dialect=dialect, **kwargs)
                for row in csv_reader:
                    encoded = []
                    for cell in row:
                        try:
                            cell = str(cell)
                        except Exception:
                            cell = str(cell.decode(ascii_codec))

                        encoded.append(cell)

                    yield encoded

            reader = unicode_csv_reader(reader)

            # read in our header
            line_number = 0

            header = next(reader)
            line_number += 1
            while header is not None and len(header[0]) > 1 and header[0][0] == "#":
                header = next(reader)
                line_number += 1

            # do some sanity checking to make sure they uploaded the right kind of file
            if len(header) < 1:
                raise Exception("Invalid header for import file")

            # normalize our header names, removing quotes and spaces
            headers = [cls.normalize_value(_).lower() for _ in header]

        return headers

    @classmethod
    def finalize_import(cls, task, records):
        """
        Provides SmartModels a hook to perform any last updates on imported records outside the import
        database transaction.
        """
        return

    @classmethod
    def import_csv(cls, task, log=None):
        csv_file = task.csv_file
        csv_file.open()

        # this file isn't good enough, lets write it to local disk
        from django.conf import settings
        from uuid import uuid4
        import os

        # make sure our tmp directory is present (throws if already present)
        try:
            os.makedirs(os.path.join(settings.MEDIA_ROOT, 'tmp'))
        except Exception:
            pass

        # write our file out
        tmp_file = os.path.join(settings.MEDIA_ROOT, 'tmp/%s' % str(uuid4()))

        out_file = open(tmp_file, 'wb')
        out_file.write(csv_file.read())
        out_file.close()

        filename = out_file
        user = task.created_by

        import_params = None
        import_results = dict()

        # additional parameters are optional
        if task.import_params:
            try:
                import_params = json.loads(task.import_params)
            except Exception:
                pass

        try:
            records = cls.import_xls(filename, user, import_params, log, import_results)
        except XLRDError:
            records = cls.import_raw_csv(filename, user, import_params, log, import_results)
        finally:
            os.remove(tmp_file)

        task.import_results = json.dumps(import_results)

        return records

    @classmethod
    def normalize_value(cls, val):
        # remove surrounding whitespace
        val = val.strip()

        # if surrounded by double quotes, remove those
        if val and val[0] == '"' and val[-1] == '"':
            val = val[1:-1]

        # if surrounded by single quotes, remove those
        if val and val[0] == "'" and val[-1] == "'":
            val = val[1:-1]

        return val

    @classmethod
    def import_xls(cls, filename, user, import_params, log=None, import_results=None):
        workbook = open_workbook(filename.name, 'rb')

        # timezone for date cells can be specified as an import parameter or defaults to UTC
        # use now to determine a relevant timezone
        naive_timezone = pytz.timezone(import_params['timezone']) \
            if import_params and 'timezone' in import_params else pytz.UTC
        tz = timezone.now().astimezone(naive_timezone).tzinfo

        records = []
        num_errors = 0
        error_messages = []

        for sheet in workbook.sheets():
            # read our header
            header = []
            for col in range(sheet.ncols):
                header.append(str(sheet.cell(0, col).value))
            header = [cls.normalize_value(_).lower() for _ in header]

            cls.validate_import_header(header)

            # read our rows
            line_number = 1
            for row in range(sheet.nrows - 1):
                field_values = []
                for col in range(sheet.ncols):
                    cell = sheet.cell(row + 1, col)
                    field_values.append(cls.get_cell_value(workbook, tz, cell))

                field_values = dict(zip(header, field_values))
                field_values['created_by'] = user
                field_values['modified_by'] = user

                try:
                    field_values = cls.prepare_fields(field_values, import_params, user)
                    record = cls.create_instance(field_values)

                    if record:
                        records.append(record)
                    else:
                        num_errors += 1

                except SmartImportRowError as e:
                    error_messages.append(dict(line=line_number+1, error=str(e)))

                except Exception as e:
                    if log:
                        traceback.print_exc(100, log)
                    raise Exception("Line %d: %s\n\n%s" % (line_number, str(e), field_values))
                line_number += 1
            # only care about the first sheet
            break

        if import_results is not None:
            import_results['records'] = len(records)
            import_results['errors'] = num_errors + len(error_messages)
            import_results['error_messages'] = error_messages

        return records

    @classmethod
    def get_cell_value(cls, workbook, tz, cell):
        if cell.ctype == XL_CELL_DATE:
            date = xldate_as_tuple(cell.value, workbook.datemode)
            return datetime.datetime(*date, tzinfo=tz)
        else:
            return cls.normalize_value(str(cell.value))

    @classmethod
    def import_raw_csv(cls, filename, user, import_params, log=None, import_results=None):
        # our alternative codec, by default we are the crazy windows encoding
        ascii_codec = 'cp1252'

        # read the entire file, look for mac_roman characters
        reader = open(filename.name, "rb")
        for byte in reader.read():
            # these are latin accented characterse in mac_roman, if we see them then our alternative
            # encoding should be mac_roman
            try:
                byte_number = ord(byte)
            except TypeError:
                byte_number = byte

            if byte_number in [0x81, 0x8d, 0x8f, 0x90, 0x9d]:
                ascii_codec = 'mac_roman'
                break
        reader.close()

        reader = open(filename.name, "rU", encoding='utf-8-sig')

        def unicode_csv_reader(utf8_data, dialect=csv.excel, **kwargs):
            csv_reader = csv.reader(utf8_data, dialect=dialect, **kwargs)
            for row in csv_reader:
                encoded = []
                for cell in row:
                    try:
                        cell = str(cell)
                    except Exception:
                        cell = str(cell.decode(ascii_codec))

                    encoded.append(cell)

                yield encoded

        reader = unicode_csv_reader(reader)

        # read in our header
        line_number = 0

        header = next(reader)
        line_number += 1
        while header is not None and len(header[0]) > 1 and header[0][0] == "#":
            header = next(reader)
            line_number += 1

        # do some sanity checking to make sure they uploaded the right kind of file
        if len(header) < 1:
            raise Exception("Invalid header for import file")

        # normalize our header names, removing quotes and spaces
        header = [cls.normalize_value(_).lower() for _ in header]

        cls.validate_import_header(header)

        records = []
        num_errors = 0
        error_messages = []

        for row in list(reader):
            # trim all our values
            row = [cls.normalize_value(_) for _ in row]

            line_number += 1

            # make sure there are same number of fields
            if len(row) != len(header):
                raise Exception("Line %d: The number of fields for this row is incorrect. Expected %d but found %d."
                                % (line_number, len(header), len(row)))

            field_values = dict(zip(header, row))
            field_values['created_by'] = user
            field_values['modified_by'] = user
            try:
                field_values = cls.prepare_fields(field_values, import_params, user)
                record = cls.create_instance(field_values)
                if record:
                    records.append(record)
                else:
                    num_errors += 1

            except SmartImportRowError as e:
                error_messages.append(dict(line=line_number, error=str(e)))

            except Exception as e:
                if log:
                    traceback.print_exc(100, log)
                raise Exception("Line %d: %s\n\n%s" % (line_number, str(e), field_values))

        if import_results is not None:
            import_results['records'] = len(records)
            import_results['errors'] = num_errors + len(error_messages)
            import_results['error_messages'] = error_messages

        return records


class ActiveManager(models.Manager):
    """
    A manager that only selects items which are still active.
    """
    def get_queryset(self):
        """
        Where the magic happens, we automatically throw on an extra is_active = True to every filter
        """
        return super(ActiveManager, self).get_queryset().filter(is_active=True)



from django.conf import settings
from django.conf.urls import url
from django.contrib.auth.views import LogoutView
from .views import Login, UserCRUDL

logout_url = getattr(settings, 'LOGOUT_REDIRECT_URL', None)

urlpatterns = [
    url(
        r'^login/$', Login.as_view(), dict(template_name='smartmin/users/login.html'),
        name="users.user_login"
    ),
    url(
        r'^logout/$', LogoutView.as_view(), dict(redirect_field_name='go', next_page=logout_url),
        name="users.user_logout"
    ),
]

urlpatterns += UserCRUDL().as_urlpatterns()



from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ImportTask',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('is_active', models.BooleanField(default=True, help_text='Whether this item is active, use this instead of deleting')),
                ('created_on', models.DateTimeField(help_text='When this item was originally created', auto_now_add=True)),
                ('modified_on', models.DateTimeField(help_text='When this item was last modified', auto_now=True)),
                ('csv_file', models.FileField(help_text='A comma delimited file of records to import', upload_to='csv_imports', verbose_name='Import file')),
                ('model_class', models.CharField(help_text='The model we are importing for', max_length=255)),
                ('import_params', models.TextField(help_text='JSON blob of form parameters on task creation', null=True, blank=True)),
                ('import_log', models.TextField()),
                ('import_results', models.TextField(help_text='JSON blob of result values on task completion', null=True, blank=True)),
                ('task_id', models.CharField(max_length=64, null=True)),
                ('created_by', models.ForeignKey(related_name='csv_imports_importtask_creations', on_delete=models.PROTECT, to=settings.AUTH_USER_MODEL, help_text='The user which originally created this item')),
                ('modified_by', models.ForeignKey(related_name='csv_imports_importtask_modifications', on_delete=models.PROTECT, to=settings.AUTH_USER_MODEL, help_text='The user which last modified this item')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]




import json
import pytz
import re
from datetime import datetime, timedelta

from django import template
from django.conf import settings
from django.urls import reverse
from django.template import TemplateSyntaxError
from django.utils import timezone

register = template.Library()


@register.simple_tag
def get_hostname():
    if settings.HOSTNAME:
        return settings.HOSTNAME
    return "localhost"


@register.simple_tag(takes_context=True)
def get_list_class(context, list):
    """
    Returns the class to use for the passed in list.  We just build something up
    from the object type for the list.
    """
    return "list_%s_%s" % (list.model._meta.app_label, list.model._meta.model_name)


def format_datetime(time):
    """
    Formats a date, converting the time to the user timezone if one is specified
    """
    user_time_zone = timezone.get_current_timezone()
    if time.tzinfo is None:
        time = time.replace(tzinfo=pytz.utc)
        user_time_zone = pytz.timezone(getattr(settings, 'USER_TIME_ZONE', 'GMT'))

    time = time.astimezone(user_time_zone)
    return time.strftime("%b %d, %Y %H:%M")


@register.simple_tag(takes_context=True)
def get_value_from_view(context, field):
    """
    Responsible for deriving the displayed value for the passed in 'field'.

    This first checks for a particular method on the ListView, then looks for a method
    on the object, then finally treats it as an attribute.
    """
    view = context['view']
    obj = None
    if 'object' in context:
        obj = context['object']

    value = view.lookup_field_value(context, obj, field)

    # it's a date
    if type(value) == datetime:
        return format_datetime(value)

    return value


@register.simple_tag(takes_context=True)
def get_value(context, obj, field):
    """
    Responsible for deriving the displayed value for the passed in 'field'.

    This first checks for a particular method on the ListView, then looks for a method
    on the object, then finally treats it as an attribute.
    """
    view = context['view']
    value = view.lookup_field_value(context, obj, field)
    if type(value) == datetime:
        return format_datetime(value)

    return value


@register.simple_tag(takes_context=True)
def get_class(context, field, obj=None):
    """
    Looks up the class for this field
    """
    view = context['view']
    return view.lookup_field_class(field, obj, "field_" + field)


@register.simple_tag(takes_context=True)
def get_label(context, field, obj=None):
    """
    Responsible for figuring out the right label for the passed in field.

    The order of precedence is:
       1) if the view has a field_config and a label specified there, use that label
       2) check for a form in the view, if it contains that field, use it's value
    """
    view = context['view']
    return view.lookup_field_label(context, field, obj)


@register.simple_tag(takes_context=True)
def get_field_link(context, field, obj=None):
    """
    Determine what the field link should be for the given field, object pair
    """
    view = context['view']
    return view.lookup_field_link(context, field, obj)


@register.simple_tag(takes_context=True)
def view_as_json(context):
    """
    Returns our view serialized as json
    """
    view = context['view']
    return json.dumps(view.as_json(context))


@register.simple_tag(takes_context=True)
def ssl_url(context, url_name, args=None):
    path = reverse(url_name, args)
    if getattr(settings, 'SESSION_COOKIE_SECURE', False):
        return "https://%s%s" % (settings.HOSTNAME, path)
    else:
        return path


@register.simple_tag(takes_context=True)
def non_ssl_url(context, url_name, args=None):
    path = reverse(url_name, args)
    if settings.HOSTNAME != "localhost":
        return "http://%s%s" % (settings.HOSTNAME, path)
    return path


@register.filter
def field(form, field):
    try:
        return form[field]
    except KeyError:
        return None


@register.filter(name='add_css')
def add_css(field, css):
    custom_attrs = field.field.widget.attrs

    if not custom_attrs.get('class', None):
        custom_attrs['class'] = css
    else:
        custom_attrs['class'] += " " + css

    return field.as_widget(attrs=custom_attrs)


@register.filter
def map(string, args):
    return string % args.__dict__


@register.filter
def gmail_time(dtime, now=None):
    if dtime.tzinfo is None:
        dtime = dtime.replace(tzinfo=pytz.utc)
        user_time_zone = pytz.timezone(getattr(settings, 'USER_TIME_ZONE', 'GMT'))
        dtime = dtime.astimezone(user_time_zone)
    else:
        dtime = dtime.astimezone(timezone.get_current_timezone())

    if not now:
        now = timezone.now()

    if now.tzinfo is None:
        now = now.replace(tzinfo=pytz.utc)

    twelve_hours_ago = now - timedelta(hours=12)

    if dtime > twelve_hours_ago:
        return "%d:%s %s" % (int(dtime.strftime("%I")), dtime.strftime("%M"), dtime.strftime("%p").lower())
    elif now.year == dtime.year:
        return "%s %d" % (dtime.strftime("%b"), int(dtime.strftime("%d")))
    else:
        return "%d/%d/%s" % (int(dtime.strftime("%d")), int(dtime.strftime("%m")), dtime.strftime("%y"))


@register.filter
def user_as_string(user):
    full_name = user.get_full_name()
    if full_name:
        return full_name
    else:
        return user.username


@register.filter
def field_help(view, field):
    """
    Returns the field help for the passed in field
    """
    return view.lookup_field_help(field)


@register.filter
def get(dictionary, key):
    """
    Simple dict lookup using two variables
    """
    if key in dictionary:
        return dictionary[key]
    else:
        return ''


@register.filter
def is_smartobject(obj):
    """
    Returns whether the passed in object is a smart object
    """
    return hasattr(obj, 'is_active')


@register.filter
def field_orderable(view, field):
    """
    Returns whether the passed in field is orderable
    """
    return view.lookup_field_orderable(field)


class PDBNode(template.Node):
    """
    Woot woot, simple pdb debugging. {% pdb %}
    """
    def render(self, context):
        import pdb; pdb.set_trace()  # noqa


@register.tag
def pdb(parser, token):
    return PDBNode()


@register.simple_tag(takes_context=True)
def getblock(context, prefix, suffix=None):
    key = prefix
    if suffix:
        key += str(suffix)

    if 'blocks' not in context:
        raise TemplateSyntaxError("setblock/endblock can only be used with SmartView or it's subclasses")

    if key in context['blocks']:
        return context['blocks'][key]
    else:
        return ""


def setblock(parser, token):
    args = token.split_contents()
    if len(args) < 2:
        raise TemplateSyntaxError("setblock tag takes one argument, the name of the block got: [%s]" % ",".join(args))

    key = "".join(args[1:])

    nodelist = parser.parse(('endsetblock',))
    parser.delete_first_token()
    return SetBlockNode(key, nodelist)


class SetBlockNode(template.Node):
    def __init__(self, key, nodelist):
        self.key = key
        self.nodelist = nodelist

    def render(self, context):
        if 'blocks' not in context:
            raise TemplateSyntaxError("setblock/endblock can only be used with SmartView or it's subclasses")

        output = self.nodelist.render(context)
        context['blocks'][self.key] = output
        return ""


# register our tag
setblock = register.tag(setblock)


@register.inclusion_tag('smartmin/field.html', takes_context=True)
def render_field(context, field):
    form = context['form']
    view = context['view']

    readonly_fields = view.derive_readonly()

    # check that this field exists in our form, either as a real field or as a readonly one
    if field not in form.fields and field not in readonly_fields:
        raise TemplateSyntaxError("Error: No field '%s' found in form to render" % field)

    inclusion_context = dict(field=field,
                             form=context['form'],
                             view=context['view'],
                             blocks=context['blocks'])
    if 'object' in context:
        inclusion_context['object'] = context['object']

    return inclusion_context


@register.simple_tag
def active(request, pattern):
    """
    Simple tag let us define a regex for the active navigation tab
    """
    if re.search(pattern, request.path):
        return 'active'
    return ''




from django.forms import widgets
from django.utils.html import escape
from django.utils.safestring import mark_safe


class VisibleHiddenWidget(widgets.Widget):

    def render(self, name, value, attrs=None, renderer=None):
        """
        Returns this Widget rendered as HTML, as a Unicode string.

        The 'value' given is not guaranteed to be valid input, so subclass
        implementations should program defensively.
        """
        html = ''
        html += '%s' % value
        html += '<input type="hidden" name="%s" value="%s">' % (escape(name), escape(value))
        return mark_safe(html)


class DatePickerWidget(widgets.DateInput):
    """
    Date input which uses Javascript date picker widget
    """
    input_format = ('MM d, yyyy', '%B %d, %Y')  # Javascript and Python format strings

    def __init__(self, *args, **kwargs):
        kwargs['attrs'] = {'data-provide': 'datepicker', 'data-date-format': self.input_format[0]}
        kwargs['format'] = self.input_format[1]

        super(DatePickerWidget, self).__init__(*args, **kwargs)

    class Media:
        js = ('js/bootstrap-datepicker.js',)
        css = {'all': ('css/bootstrap-datepicker3.css',)}


class ImageThumbnailWidget(widgets.ClearableFileInput):

    def __init__(self, thumb_width=75, thumb_height=75):
        self.width = thumb_width
        self.height = thumb_height
        super(ImageThumbnailWidget, self).__init__({})

    def render(self, name, value, attrs=None, renderer=None):
        thumb_html = '<table><tr>'
        if value and hasattr(value, "url"):
            try:
                from sorl.thumbnail import get_thumbnail
                value = get_thumbnail(value, f"{self.width}x{self.height}", crop='center', quality=99)
            except ImportError:
                pass

            thumb_html += '<td><img src="%s" width="%s" width="%s" /></td>' % (value.url, self.width, self.height)

        thumb_html += '<td><input type="checkbox" name="%s-clear" /> Clear' % name
        thumb_html += '<input type="file" name="%s" /></td>' % name
        thumb_html += '</tr></table>'

        return mark_safe(str('<div class="image-picker">%s</div>' % thumb_html))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smartmin',
 'smartmin.csv_imports',
 'smartmin.csv_imports.migrations',
 'smartmin.management',
 'smartmin.management.commands',
 'smartmin.templatetags',
 'smartmin.users',
 'smartmin.users.migrations']

package_data = \
{'': ['*'],
 'smartmin': ['static/*',
              'static/css/*',
              'static/fonts/*',
              'static/img/*',
              'static/img/smartmin/*',
              'static/js/*',
              'static/js/libs/*',
              'templates/*',
              'templates/csv_imports/*',
              'templates/smartmin/*',
              'templates/smartmin/users/*']}

install_requires = \
['Django>=2.2,<4.0',
 'celery>=4.4.6',
 'pytz',
 'redis>=3.5.3,<4.0.0',
 'sqlparse>=0.4.1,<0.5.0',
 'xlrd>=1.2.0,<2.0.0',
 'xlwt>=1.3.0,<2.0.0']

setup_kwargs = {
    'name': 'smartmin',
    'version': '3.1.0',
    'description': 'Scaffolding system for Django object management.',
    'long_description': "Django Smartmin\n===============\n\n[![Build Status](https://github.com/nyaruka/smartmin/workflows/CI/badge.svg)](https://github.com/nyaruka/smartmin/actions?query=workflow%3ACI) \n[![codecov](https://codecov.io/gh/nyaruka/smartmin/branch/main/graph/badge.svg)](https://codecov.io/gh/nyaruka/smartmin)\n[![PyPI Release](https://img.shields.io/pypi/v/smartmin.svg)](https://pypi.python.org/pypi/smartmin/)\n\nSmartmin was born out of the frustration of the Django admin site not being well suited to being exposed to clients. \nIt aims to allow you to quickly build scaffolding which you can customize by using Django views.\n\nIt is very opinionated in how it works, if you don't agree, Smartmin may not be for you:\n\n- Permissions are used to gate access to each page, embrace permissions throughout and you'll love this\n- CRUDL operations at the object level, that is, Create, Read, Update, Delete and List, permissions and views are based \n  around this\n- URL automapping via the the CRUDL objects, this should keep things very very DRY\n\nAbout Versions\n==============\n\nSmartmin tries to stay in lock step with the latest Django versions. With each new Django LTS release we will release\na new Smartmin major version and we will reserve major changes (possibly breaking backwards compatibility) for such \nreleases. This includes updating to the latest version of Twitter Bootstrap.\n\nThe latest version is the 3.* series which supports the Django 3.2 LTS and 2.2 LTS.\n\nAbout\n=====\n\nThe full documentation can be found at: http://readthedocs.org/docs/smartmin/en/latest/\n\nThe official source code repository is: http://www.github.com/nyaruka/smartmin/\n\nBuilt in Rwanda by [Nyaruka Ltd](http://www.nyaruka.com).\n",
    'author': 'Nyaruka Ltd',
    'author_email': 'code@nyaruka.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import re
from datetime import timedelta

from django.conf import settings
from django.contrib.auth.hashers import check_password
from django.db import models
from django.utils import timezone


def is_password_complex(password):
    has_caps = re.search('[A-Z]+', password)
    has_lower = re.search('[a-z]+', password)
    has_digit = re.search('[0-9]+', password)

    if len(password) < 8 or (len(password) < 12 and (not has_caps or not has_lower or not has_digit)):
        return False
    else:
        return True


class RecoveryToken(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    token = models.CharField(max_length=32, unique=True, default=None, help_text="token to reset password")
    created_on = models.DateTimeField(auto_now_add=True)


class FailedLogin(models.Model):
    username = models.CharField(max_length=256)
    failed_on = models.DateTimeField(auto_now_add=True)


class PasswordHistory(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
                             on_delete=models.PROTECT,
                             help_text="The user that set a password")
    password = models.CharField(max_length=255,
                                help_text="The hash of the password that was set")
    set_on = models.DateTimeField(auto_now_add=True,
                                  help_text="When the password was set")

    @classmethod
    def is_password_repeat(cls, user, password):
        password_window = getattr(settings, 'USER_PASSWORD_REPEAT_WINDOW', -1)
        if password_window <= 0:
            return False

        # check their current password
        if check_password(password, user.password):
            return True

        # get all the passwords in the past year
        window_ago = timezone.now() - timedelta(days=password_window)
        previous_passwords = PasswordHistory.objects.filter(user=user, set_on__gte=window_ago)
        for previous in previous_passwords:
            if check_password(password, previous.password):
                return True

        return False

    @classmethod
    def is_password_expired(cls, user):
        password_expiration = getattr(settings, 'USER_PASSWORD_EXPIRATION', -1)

        if password_expiration <= 0:
            return False

        # get the most recent password change
        last_password = PasswordHistory.objects.filter(user=user).order_by('-set_on')

        last_set = user.date_joined
        if last_password:
            last_set = last_password[0].set_on

        # calculate how long ago our password was set
        today = timezone.now()
        difference = (today - last_set)

        # return whether that is expired
        return difference.days > password_expiration



from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='FailedLogin',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('failed_on', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, on_delete=models.PROTECT)),
            ],
        ),
        migrations.CreateModel(
            name='PasswordHistory',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('password', models.CharField(help_text='The hash of the password that was set', max_length=255)),
                ('set_on', models.DateTimeField(help_text='When the password was set', auto_now_add=True)),
                ('user', models.ForeignKey(help_text='The user that set a password', on_delete=models.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='RecoveryToken',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('token', models.CharField(default=None, help_text='token to reset password', unique=True, max_length=32)),
                ('created_on', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, on_delete=models.PROTECT)),
            ],
        ),
    ]



from django.utils.module_loading import import_string

from smartmin.csv_imports.models import ImportTask
from smartmin.views import SmartCRUDL, SmartListView, SmartReadView


class ImportTaskCRUDL(SmartCRUDL):
    model = ImportTask
    actions = ('read', 'list')

    class Read(SmartReadView):
        def derive_refresh(self):
            if self.object.status() in ["PENDING", "RUNNING", "STARTED"]:
                return 2000
            else:
                return 0

    class List(SmartListView):
        fields = ('status', 'type', 'csv_file', 'created_on', 'created_by')
        link_fields = ('csv_file',)

        def get_type(self, obj):
            return import_string(obj.model_class)._meta.verbose_name_plural.title()
